from __future__ import annotations

import asyncio
from datetime import datetime, time
from functools import lru_cache
import inspect
import pkgutil
import random
import re
from typing import List, Type
from importlib import import_module
from pathlib import Path
import json

from loguru import logger

from ..utils import next_random_datetime
from . import __name__ as __product__
from .link import Link
from .tele import ClientsSession

from .bots._base import BaseBotCheckin, CheckinResult

logger = logger.bind(scheme="telegram")


def get_spec(type: str):
    """服务模块路径解析."""
    if type == "checkiner":
        sub = "bots"
        suffix = "checkin"
    elif type == "monitor":
        sub = "monitor"
        suffix = "monitor"
    elif type == "messager":
        sub = "messager"
        suffix = "messager"
    else:
        raise ValueError(f"{type} is not a valid service.")
    return sub, suffix


@lru_cache
def get_names(type: str, allow_ignore=False) -> List[str]:
    """列出服务中所有可用站点."""
    sub, _ = get_spec(type)
    results = []
    typemodule = import_module(f"{__product__}.{sub}")
    for _, mn, _ in pkgutil.iter_modules(typemodule.__path__):
        module = import_module(f"{__product__}.{sub}.{mn}")
        if not allow_ignore:
            if not getattr(module, "__ignore__", False):
                results.append(mn)
        else:
            if (not mn.startswith("_")) and (not mn.startswith("test")):
                results.append(mn)
    return results


def get_cls(type: str, names: List[str] = None) -> List[Type]:
    """获得服务特定站点的所有类."""
    sub, suffix = get_spec(type)
    if names == None:
        names = get_names(type)

    exclude_names = set(name[1:] for name in names if name.startswith("-"))
    include_names = set(name[1:] for name in names if name.startswith("+"))
    names = set(name for name in names if not name.startswith("-") and not name.startswith("+"))

    if not names and (exclude_names or include_names):
        names = set(get_names(type))

    if "all" in names:
        names = set(get_names(type, allow_ignore=True))

    if type == "checkiner":
        if "sgk" in names:
            sgk_names = set(n for n in get_names(type, allow_ignore=True) if n.endswith("sgk"))
            names.update(sgk_names)
            names.remove("sgk")

        if "sgk" in exclude_names:
            sgk_names = set(n for n in names if n.endswith("sgk"))
            names -= sgk_names
            exclude_names.remove("sgk")

        if "sgk" in include_names:
            sgk_names = set(n for n in get_names(type, allow_ignore=True) if n.endswith("sgk"))
            include_names.update(sgk_names)
            include_names.remove("sgk")

    # 应用排除项
    names = names - exclude_names
    # 添加附加项
    names = list(names | include_names)

    results = []
    for name in names:
        match = re.match(r"templ_(\w+)<(\w+)>", name)
        if match:
            try:
                module = import_module(f"{__product__}.{sub}._templ_{match.group(1).lower()}")
                func = getattr(module, "use", None)
                if not func:
                    logger.warning(f'您配置的 "{type}" 不支持模板 "{match.group(1)}".')
                    continue
                results.append(func(bot_username=match.group(2), name=f"@{match.group(2)}"))
            except ImportError:
                all_names = get_names(type)
                logger.warning(f'您配置的 "{type}" 不支持站点 "{name}", 请从以下站点中选择:')
                logger.warning(", ".join(all_names))
        else:
            try:
                module = import_module(f"{__product__}.{sub}.{name.lower()}")
                for cn, cls in inspect.getmembers(module, inspect.isclass):
                    if (name.replace("_", "").replace("_old", "") + suffix).lower() == cn.lower():
                        results.append(cls)
            except ImportError:
                all_names = get_names(type)
                logger.warning(f'您配置的 "{type}" 不支持站点 "{name}", 请从以下站点中选择:')
                logger.warning(", ".join(all_names))
    return results


def extract(clss: List[Type]) -> List[Type]:
    """对于嵌套类, 展开所有子类."""
    extracted = []
    for cls in clss:
        ncs = [c for c in cls.__dict__.values() if inspect.isclass(c)]
        if ncs:
            extracted.extend(ncs)
        else:
            extracted.append(cls)
    return extracted


async def _checkin_task(checkiner: BaseBotCheckin, sem, wait=0):
    """签到器壳, 用于随机等待开始."""
    if wait > 0:
        checkiner.log.debug(f"随机启动等待: 将等待 {wait:.2f} 分钟以启动.")
    await asyncio.sleep(wait * 60)
    async with sem:
        result = await checkiner._start()
        await asyncio.sleep(random.uniform(5, 10))
        return result


async def _gather_task(tasks, username):
    return username, await asyncio.gather(*tasks)


async def checkiner(config: dict, instant=False):
    """签到器入口函数."""
    logger.debug("正在启动每日签到模块, 请等待登录.")
    async with ClientsSession.from_config(config, checkin=(True, True)) as clients:
        coros = []
        async for tg in clients:
            log = logger.bind(scheme="telechecker", username=tg.me.name)
            logger.info("已连接到 Telegram, 签到器正在初始化.")
            service_config = config.get("telegram", [])[tg._config_index].get("service", {})
            if not service_config:
                service_config: dict = config.get("service", {})
            names = service_config.get("checkiner", None)
            clses = extract(get_cls("checkiner", names=names))
            if not clses:
                log.warning("没有任何有效签到站点, 签到将跳过.")
                continue
            if not await Link(tg).auth("checkiner", log_func=log.error):
                continue
            sem = asyncio.Semaphore(int(config.get("concurrent", 1)))
            checkiners: List[BaseBotCheckin] = [
                cls(
                    tg,
                    retries=config.get("retries", 4),
                    timeout=config.get("timeout", 240),
                    nofail=config.get("nofail", True),
                    basedir=config.get("basedir", None),
                    proxy=config.get("proxy", None),
                    config=config.get("checkiner", {}).get(cls.__module__.rsplit(".", 1)[-1], {}),
                    instant=instant,
                )
                for cls in clses
            ]
            tasks = []
            names = []
            for c in checkiners:
                names.append(c.name)
                wait = 0 if instant else random.uniform(0, int(config.get("random", 60)))
                task = asyncio.create_task(_checkin_task(c, sem, wait))
                tasks.append(task)
            coros.append(asyncio.ensure_future(_gather_task(tasks, username=tg.me.name)))
            if names:
                log.debug(f'已启用签到器: {", ".join(names)}')
        while coros:
            done, coros = await asyncio.wait(coros, return_when=asyncio.FIRST_COMPLETED)
            for t in done:
                try:
                    username, results = await t
                except asyncio.CancelledError:
                    continue
                log = logger.bind(scheme="telechecker", username=username)
                failed = []
                ignored = []
                successful = []
                checked = []
                for n, s in results:
                    if s == CheckinResult.IGNORE:
                        ignored.append(n)
                    elif s == CheckinResult.SUCCESS:
                        successful.append(n)
                    elif s == CheckinResult.CHECKED:
                        checked.append(n)
                    else:
                        failed.append(n)
                spec = f"共{len(checkiners)}个"
                if successful:
                    spec += f", {len(successful)}成功"
                if checked:
                    spec += f", {len(checked)}已签到而跳过"
                if failed:
                    spec += f", {len(failed)}失败"
                if ignored:
                    spec += f", {len(ignored)}跳过"
                if failed:
                    if successful:
                        msg = "签到部分失败"
                    else:
                        msg = "签到失败"
                    log.error(f"{msg} ({spec}): {', '.join([f for f in failed])}")
                else:
                    log.bind(log=True).info(f"签到成功 ({spec}).")


async def checkiner_schedule(
    config: dict,
    start_time: time = None,
    end_time: time = None,
    days: int = 1,
    instant: bool = False,
):
    """签到器计划任务."""

    timestamp_file = Path(config["basedir"]) / "checkiner_schedule_next_timestamp"
    current_config = {
        "start_time": start_time.strftime("%H:%M") if start_time else None,
        "end_time": end_time.strftime("%H:%M") if end_time else None,
        "days": days,
    }

    while True:
        next_dt = None
        config_changed = False

        if timestamp_file.exists():
            try:
                stored_data = json.loads(timestamp_file.read_text())
                if not isinstance(stored_data, dict):
                    raise ValueError("invalid cache")
                stored_timestamp = stored_data["timestamp"]
                stored_config = stored_data["config"]

                if stored_config != current_config:
                    logger.bind(scheme="telechecker").info("计划任务配置已更改，将重新计算下次执行时间.")
                    config_changed = True
                else:
                    next_dt = datetime.fromtimestamp(stored_timestamp)
                    if next_dt > datetime.now():
                        logger.bind(scheme="telechecker").info(
                            f"从缓存中读取到下次签到时间: {next_dt.strftime('%m-%d %H:%M %p')}."
                        )
            except (ValueError, OSError, json.JSONDecodeError) as e:
                logger.debug(f"读取存储的时间戳失败: {e}")
                config_changed = True

        if not next_dt or next_dt <= datetime.now() or config_changed:
            next_dt = next_random_datetime(start_time, end_time, interval_days=days)
            logger.bind(scheme="telechecker").info(
                f"下一次签到将在 {next_dt.strftime('%m-%d %H:%M %p')} 进行."
            )
            try:
                save_data = {"timestamp": next_dt.timestamp(), "config": current_config}
                timestamp_file.write_text(json.dumps(save_data))
            except OSError as e:
                logger.debug(f"存储时间戳失败: {e}")

        await asyncio.sleep((next_dt - datetime.now()).total_seconds())

        try:
            timestamp_file.unlink(missing_ok=True)
        except OSError as e:
            logger.debug(f"删除时间戳文件失败: {e}")

        await checkiner(config, instant=instant)


async def monitorer(config: dict):
    """监控器入口函数."""
    logger.debug("正在启动消息监控模块.")
    jobs = []
    async with ClientsSession.from_config(config, monitor=(True, False)) as clients:
        async for tg in clients:
            log = logger.bind(scheme="telemonitor", username=tg.me.name)
            logger.info("已连接到 Telegram, 监控器正在初始化.")
            service_config = config.get("telegram", [])[tg._config_index].get("service", {})
            if not service_config:
                service_config: dict = config.get("service", {})
            names = service_config.get("monitor", None)
            clses = extract(get_cls("monitor", names=names))
            if not clses:
                log.warning("没有任何有效监控站点, 监控将跳过.")
            if not await Link(tg).auth("monitorer", log_func=log.error):
                continue
            names = []
            for cls in clses:
                cls_config = config.get("monitor", {}).get(cls.__module__.rsplit(".", 1)[-1], {})
                jobs.append(
                    asyncio.create_task(
                        cls(
                            tg,
                            nofail=config.get("nofail", True),
                            basedir=config.get("basedir", None),
                            proxy=config.get("proxy", None),
                            config=cls_config,
                        )._start()
                    )
                )
                names.append(cls.name)
            if names:
                log.debug(f'已启用监控器: {", ".join(names)}')
        await asyncio.gather(*jobs)


async def messager(config: dict):
    """自动水群入口函数."""
    logger.debug("正在启动自动水群模块.")
    messagers = []
    async with ClientsSession.from_config(config, send=(True, False)) as clients:
        async for tg in clients:
            log = logger.bind(scheme="telemessager", username=tg.me.name)
            logger.info("已连接到 Telegram, 自动水群正在初始化.")
            service_config = config.get("telegram", [])[tg._config_index].get("service", {})
            if not service_config:
                service_config: dict = config.get("service", {})
            names = service_config.get("messager", None)
            clses = extract(get_cls("messager", names=names))
            if not clses:
                log.warning("没有任何有效自动水群站点, 自动水群将跳过.")
            if not await Link(tg).auth("messager", log_func=log.error):
                continue
            for cls in clses:
                cls_config = config.get("messager", {}).get(cls.__module__.rsplit(".", 1)[-1], {})
                messagers.append(
                    cls(
                        {"api_id": tg.api_id, "api_hash": tg.api_hash, "phone": tg.phone_number},
                        me=tg.me,
                        nofail=config.get("nofail", True),
                        proxy=config.get("proxy", None),
                        basedir=config.get("basedir", None),
                        config=cls_config,
                    )
                )
    await asyncio.gather(*[m._start() for m in messagers])



import asyncio
import random
import re
from pyrogram.types import Message
from pyrogram.raw.types.messages import BotCallbackAnswer

from ._base import BotCheckin


class PilipiliCheckin(BotCheckin):
    name = "Pilipili"
    bot_username = "PiliPiliUltraTv_bot"
    bot_checkin_caption_pat = "<disabled>"
    bot_checkin_cmd = "/start"
    additional_auth = ["prime"]

    async def message_handler(self, client, message: Message):
        if (
            message.caption
            and ("请选择功能" in message.caption or "用户面板" in message.caption)
            and message.reply_markup
        ):
            keys = [k.text for r in message.reply_markup.inline_keyboard for k in r]
            for k in keys:
                if "签到" in k:
                    try:
                        answer: BotCallbackAnswer = await message.click(k)
                        await self.on_text(Message(id=0), answer.message)
                    except TimeoutError:
                        pass
                    return
            else:
                self.log.warning(f"签到失败: 账户错误.")
                return await self.fail()

        if message.caption and "签到说明" in message.caption:
            text = message.caption.replace("×", "*").replace("÷", "/")
            match = re.search(r"计算出\s*(\d+)\s*([+\-*/])\s*(\d+)\s*=\s*\?", text)
            if match:
                num1, operator, num2 = match.groups()
                num1, num2 = int(num1), int(num2)
                if operator == "+":
                    result = int(num1 + num2)
                elif operator == "-":
                    result = int(num1 - num2)
                elif operator == "*":
                    result = int(num1 * num2)
                elif operator == "/":
                    result = int(num1 / num2)
                self.log.info(f"解析数学题答案: {num1}{operator}{num2}={result}")
                await asyncio.sleep(random.uniform(10, 15))
                await self.client.send_message(self.bot_username, str(result))
                return
            else:
                self.log.warning(f"签到时出现未知题目.")

        if message.text and "请先点击下面加入我们的" in message.text:
            self.log.warning(f"签到失败: 账户错误.")
            return await self.fail()

        await super().message_handler(client, message)



from __future__ import annotations

import asyncio
from pathlib import Path
import random
import string
from typing import TYPE_CHECKING, Iterable, Tuple, Union
from datetime import datetime, time, timezone
import warnings
import json

import httpx
from loguru import logger
from dateutil import parser

with warnings.catch_warnings():
    warnings.filterwarnings("ignore", category=DeprecationWarning)
    from embypy.objects import Episode, Movie

from ..utils import show_exception, next_random_datetime, truncate_str
from ..var import debug
from .emby import Emby, Connector, EmbyObject

if TYPE_CHECKING:
    from loguru import Logger

logger = logger.bind(scheme="embywatcher")


class PlayError(Exception):
    pass


def is_ok(co):
    """判定返回来自 emby 的响应为成功."""
    if isinstance(co, tuple):
        co, *_ = co
    if 200 <= co < 300:
        return True


async def get_random_media(emby: Emby):
    """获取随机视频."""
    while True:
        items = await emby.get_items(["Movie", "Episode"], limit=3, sort="Random", ascending=False)
        i: Union[Movie, Episode]
        for i in items:
            yield i


async def set_played(obj: EmbyObject):
    """设定已播放."""
    c: Connector = obj.connector
    return is_ok(await c.post(f"/Users/{{UserId}}/PlayedItems/{obj.id}"))


async def hide_from_resume(obj: EmbyObject):
    """从首页的"继续收看"部分隐藏."""
    c: Connector = obj.connector
    try:
        return is_ok(await c.post(f"/Users/{{UserId}}/Items/{obj.id}/HideFromResume", Hide=True))
    except RuntimeError:
        return False


def get_last_played(obj: EmbyObject):
    """获取上次播放时间."""
    last_played = obj.object_dict.get("UserData", {}).get("LastPlayedDate", None)
    return parser.parse(last_played) if last_played else None


async def play(obj: EmbyObject, loggeruser: Logger, time: float = 10):
    """模拟播放视频."""
    c: Connector = obj.connector

    await asyncio.sleep(random.uniform(1, 3))

    # 获取页面信息
    resp = await c.getJson(
        f"/Videos/{obj.id}/AdditionalParts",
        Fields="PrimaryImageAspectRatio,UserData,CanDelete",
        IncludeItemTypes="Playlist,BoxSet",
        Recursive=True,
        SortBy="SortName",
    )

    await asyncio.sleep(random.uniform(1, 3))

    # 获取播放源
    resp = await c.postJson(
        f"/Items/{obj.id}/PlaybackInfo",
        UserID=c.userid,
        StartTimeTicks=0,
        IsPlayback=True,
        AutoOpenLiveStream=True,
        MaxStreamingBitrate=42000000,
    )
    play_session_id = resp.get("PlaySessionId", "")
    if "MediaSources" in resp:
        media_source_id = resp["MediaSources"][0]["Id"]
        direct_stream_id = resp["MediaSources"][0].get("DirectStreamUrl", None)
    else:
        media_source_id = "".join(random.choice(string.ascii_lowercase + string.digits) for _ in range(32))
        direct_stream_id = None

    await asyncio.sleep(random.uniform(1, 3))

    # 模拟播放
    playing_info = lambda tick: {
        "VolumeLevel": 100,
        "CanSeek": True,
        "BufferedRanges": [{"start": 0, "end": tick}] if tick else [],
        "IsPaused": False,
        "ItemId": obj.id,
        "MediaSourceId": media_source_id,
        "PlayMethod": "DirectStream",
        "PlaySessionId": play_session_id,
        "PlaylistIndex": 0,
        "PlaylistLength": 1,
        "PositionTicks": tick,
        "RepeatMode": "RepeatNone",
    }

    task = asyncio.create_task(c.get_stream_noreturn(direct_stream_id or f"/Videos/{obj.id}/stream"))

    try:
        await asyncio.sleep(random.uniform(1, 3))

        resp = await c.post("Sessions/Playing", MediaSourceId=media_source_id, data=playing_info(0))

        if not is_ok(resp):
            raise PlayError("无法开始播放")

        t = time
        last_report_t = t
        progress_errors = 0
        while t > 0:
            if progress_errors > 12:
                raise PlayError("播放状态设定错误次数过多")
            if last_report_t and last_report_t - t > (5 if debug else 30):
                loggeruser.info(f'正在播放: "{truncate_str(obj.name, 10)}" (还剩 {t:.0f} 秒).')
                last_report_t = t
            st = random.uniform(2, 5)
            await asyncio.sleep(st)
            t -= st
            tick = int((time - t) * 10000000)
            payload = playing_info(tick)
            try:
                resp = await asyncio.wait_for(
                    c.post("/Sessions/Playing/Progress", data=payload, EventName="timeupdate"), 10
                )
            except (httpx.HTTPError, asyncio.TimeoutError) as e:
                loggeruser.debug(f"播放状态设定错误: {e}")
                progress_errors += 1
            else:
                if not is_ok(resp):
                    loggeruser.debug(f"播放状态设定错误: {resp}")
                    progress_errors += 1

        await asyncio.sleep(random.uniform(1, 3))
    finally:
        task.cancel()
        try:
            await task
        except asyncio.CancelledError:
            pass
        except Exception as e:
            loggeruser.warning(f"模拟播放时, 访问流媒体文件失败.")
            show_exception(e)

    for retry in range(3):
        try:
            if not is_ok(await c.post("/Sessions/Playing/Stopped", data=playing_info(time * 10000000))):
                if retry == 2:
                    raise PlayError("尝试停止播放3次后仍然失败")
                loggeruser.debug(f"停止播放失败，正在进行第 {retry + 1}/3 次尝试")
                await asyncio.sleep(1)
                continue
            else:
                loggeruser.info(f"播放完成, 共 {time:.0f} 秒.")
                return True
        except httpx.HTTPError as e:
            if retry == 2:
                raise PlayError(f"由于连接错误或服务器错误无法停止播放: {e}")
            loggeruser.debug(f"停止播放时发生连接错误或服务器错误，正在进行第 {retry + 1}/3 次尝试: {e}")
            await asyncio.sleep(1)


async def get_cf_clearance(config, url):
    from embykeeper.telechecker.link import Link
    from embykeeper.telechecker.tele import ClientsSession

    server_info_url = f"{url.rstrip('/')}/System/Info/Public"
    telegrams = config.get("telegram", [])
    if not len(telegrams):
        logger.warning(f"未设置 Telegram 账号, 无法为 Emby 站点使用验证码解析.")
    async with ClientsSession.from_config(config) as clients:
        async for tg in clients:
            cf_clearance, proxy = await Link(tg).captcha_emby(server_info_url)
            return cf_clearance, proxy
    return None, None


async def login(config, continuous=False):
    """登录账号."""
    for a in config.get("emby", ()):
        if not continuous == a.get("continuous", False):
            continue
        logger.info(f'登录账号: "{a["username"]}" 至服务器: "{a["url"]}"')

        info = None
        for _ in range(3):
            cf_clearance = None
            proxy = None

            device_id = a.get("device_id", None)
            if not device_id:
                device_id_file = Path(config["basedir"]) / "emby_device_id"
                if device_id_file.exists():
                    try:
                        device_id = device_id_file.read_text().strip()
                    except OSError as e:
                        logger.debug(f"读取 device_id 文件失败: {e}")
                if not device_id:
                    from .emby import Connector

                    device_id = str(Connector.get_device_uuid()).upper()
                    try:
                        device_id_file.write_text(device_id)
                    except OSError as e:
                        logger.debug(f"保存 device_id 文件失败: {e}")

            emby = Emby(
                url=a["url"],
                username=a["username"],
                password=a["password"],
                jellyfin=a.get("jellyfin", False),
                proxy=proxy or config.get("proxy", None),
                ua=a.get("ua", None),
                device=a.get("device", None),
                client=a.get("client", None),
                device_id=device_id,
                cf_clearance=cf_clearance,
            )
            try:
                info = await emby.info()
            except httpx.HTTPError as e:
                logger.error(f'Emby ({a["url"]}) 连接错误或服务器错误, 请重新检查配置: {e}')
                break
            except RuntimeError as e:
                if "Unexpected JSON output" in str(e):
                    if "cf-wrapper" in str(e) or "Enable JavaScript and cookies to continue" in str(e):
                        if a.get("cf_challenge", False):
                            logger.info(
                                f'Emby "{a["url"]}" 已启用 Cloudflare 保护, 即将请求解析 (最大支持时长 15 分钟).'
                            )
                            cf_clearance, proxy = await get_cf_clearance(config, a["url"])
                            if not cf_clearance:
                                logger.warning(f'Emby "{a["url"]}" 验证码解析失败而跳过.')
                                break
                        else:
                            logger.warning(
                                f'Emby "{a["url"]}" 已启用 Cloudflare 保护, 请使用 "cf_challenge" 配置项以允许尝试解析验证码.'
                            )
                            break
                    else:
                        raise
                else:
                    raise
            else:
                break
        else:
            logger.warning(f'Emby "{a["url"]}" 验证码解析次数过多而跳过.')

        if info:
            loggeruser = logger.bind(server=info["ServerName"], username=a["username"])
            loggeruser.info(
                f'成功连接至服务器 "{a["url"]}" ({"Jellyfin" if a.get("jellyfin", False) else "Emby"} {info["Version"]}).'
            )
            yield (
                emby,
                loggeruser,
                a.get("time", None if continuous else [120, 240]),
                True if continuous else a.get("allow_multiple", True),
                a.get("allow_stream", False),
            )
        else:
            logger.bind(log=True).error(f'Emby "{a["url"]}" 无法获取元信息而跳过, 请重新检查配置.')
            continue


async def watch(
    emby: Emby,
    loggeruser: Logger,
    time: Union[float, Tuple[float, float]],
    stream: bool = False,
    retries: int = 5,
):
    """
    主执行函数 - 观看一个视频.
    参数:
        emby: Emby 客户端
        time: 模拟播放时间
        progress: 播放后设定的观看进度
        loggeruser: 日志器
        retries: 最大重试次数
    """
    retry = 0
    while True:
        try:
            async for obj in get_random_media(emby):
                if isinstance(time, int) and time <= 0:
                    loggeruser.info(f"需要播放的时间小于0, 仅登陆.")
                    return True
                if isinstance(time, Iterable):
                    t = random.uniform(*time) + 10
                else:
                    t = time + 10
                total_ticks = obj.object_dict.get("RunTimeTicks")
                if not total_ticks:
                    if not stream:
                        rt = random.uniform(30, 60)
                        loggeruser.info(
                            f'无法获取视频 "{truncate_str(obj.name, 10)}" 长度, 等待 {rt:.0f} 秒后重试.'
                        )
                        await asyncio.sleep(rt)
                        continue
                loggeruser.info(f'开始尝试播放 "{truncate_str(obj.name, 10)}" ({t:.0f} 秒).')
                while True:
                    try:
                        await play(obj, loggeruser, time=t)

                        obj = await obj.update("UserData")

                        if obj.play_count < 1:
                            raise PlayError("尝试播放后播放数低于1")

                        last_played = get_last_played(obj)
                        if not last_played:
                            raise PlayError("尝试播放后无记录")

                        last_played = (
                            last_played.replace(tzinfo=timezone.utc)
                            .astimezone(tz=None)
                            .strftime("%m-%d %H:%M")
                        )

                        prompt = (
                            f"[yellow]成功播放视频[/], "
                            + f"当前该视频播放 {obj.play_count} 次, "
                            + f"上次播放于 {last_played} UTC."
                        )

                        loggeruser.bind(log=True).info(prompt)
                        return True
                    except httpx.HTTPError as e:
                        retry += 1
                        if retry > retries:
                            loggeruser.warning(f"超过最大重试次数, 保活失败: {e}.")
                            return False
                        else:
                            rt = random.uniform(30, 60)
                            loggeruser.info(f"连接失败或服务器错误, 等待 {rt:.0f} 秒后重试: {e}.")
                            await asyncio.sleep(rt)
                    except PlayError as e:
                        retry += 1
                        if retry > retries:
                            loggeruser.warning(f"超过最大重试次数, 保活失败: {e}.")
                            return False
                        else:
                            rt = random.uniform(30, 60)
                            loggeruser.info(f"发生错误, 等待 {rt:.0f} 秒后重试其他视频: {e}.")
                            await asyncio.sleep(rt)
                        break
                    finally:
                        try:
                            if not await asyncio.shield(asyncio.wait_for(hide_from_resume(obj), 5)):
                                loggeruser.debug(f"未能成功从最近播放中隐藏视频.")
                            else:
                                loggeruser.info(f"已从最近播放中隐藏该视频.")
                        except asyncio.TimeoutError:
                            loggeruser.debug(f"从最近播放中隐藏视频超时.")
            else:
                loggeruser.warning(f"由于没有成功播放视频, 保活失败, 请重新检查配置.")
                return False
        except httpx.HTTPError as e:
            retry += 1
            if retry > retries:
                loggeruser.warning(f"超过最大重试次数, 保活失败: {e}.")
                return False
            else:
                rt = random.uniform(30, 60)
                loggeruser.info(f"连接失败, 等待 {rt:.0f} 秒后重试: {e}.")
                await asyncio.sleep(rt)
        except asyncio.CancelledError:
            raise
        except Exception as e:
            loggeruser.warning(f"发生错误, 保活失败.")
            show_exception(e, regular=False)
            return False


async def watch_multiple(
    emby: Emby,
    loggeruser: Logger,
    time: Union[float, Tuple[float, float]],
    stream: bool = False,
    retries: int = 5,
):
    if isinstance(time, Iterable):
        req_time = random.uniform(*time) + 10
    else:
        req_time = time + 10
    if not (isinstance(time, int) and time <= 0):
        loggeruser.info(f"开始播放视频 (允许播放多个), 共需播放 {req_time:.0f} 秒.")
    played_time = 0
    played_videos = 0
    retry = 0
    while True:
        try:
            async for obj in get_random_media(emby):
                if isinstance(time, int) and time <= 0:
                    loggeruser.info(f"需要播放的时间小于0, 仅登陆.")
                    return True
                total_ticks = obj.object_dict.get("RunTimeTicks")
                if not total_ticks:
                    if stream:
                        total_ticks = min(req_time, random.randint(480, 720)) * 10000000
                    else:
                        rt = random.uniform(30, 60)
                        loggeruser.info(
                            f'无法获取视频 "{truncate_str(obj.name, 10)}" 长度, 等待 {rt:.0f} 秒后重试.'
                        )
                        await asyncio.sleep(rt)
                        continue
                total_time = total_ticks / 10000000
                if req_time - played_time > total_time:
                    play_time = total_time
                else:
                    play_time = max(req_time - played_time, 10)
                loggeruser.info(f'开始尝试播放 "{truncate_str(obj.name, 10)}" ({play_time:.0f} 秒).')
                while True:
                    try:
                        await play(obj, loggeruser, time=play_time)

                        obj = await obj.update("UserData")

                        if obj.play_count < 1:
                            raise PlayError("尝试播放后播放数低于1")

                        last_played = get_last_played(obj)
                        if not last_played:
                            raise PlayError("尝试播放后无记录")

                        last_played = (
                            last_played.replace(tzinfo=timezone.utc)
                            .astimezone(tz=None)
                            .strftime("%m-%d %H:%M")
                        )

                        prompt = (
                            f"[yellow]成功播放视频[/], "
                            + f"当前该视频播放 {obj.play_count} 次, "
                            + f"上次播放于 {last_played} UTC."
                        )
                        loggeruser.info(prompt)
                        played_videos += 1
                        played_time += play_time

                        if played_time >= req_time - 1:
                            loggeruser.bind(log=True).info(f"保活成功, 共播放 {played_videos} 个视频.")
                            return True
                        else:
                            loggeruser.info(f"还需播放 {req_time - played_time:.0f} 秒.")
                            rt = random.uniform(5, 15)
                            loggeruser.info(f"等待 {rt:.0f} 秒后播放下一个.")
                            await asyncio.sleep(rt)
                            break
                    except httpx.HTTPError as e:
                        retry += 1
                        if retry > retries:
                            loggeruser.warning(f"超过最大重试次数, 保活失败: {e}.")
                            return False
                        else:
                            rt = random.uniform(30, 60)
                            loggeruser.info(f"连接失败, 等待 {rt:.0f} 秒后重试: {e}.")
                            await asyncio.sleep(rt)
                    except PlayError as e:
                        retry += 1
                        if retry > retries:
                            loggeruser.warning(f"超过最大重试次数, 保活失败: {e}.")
                            return False
                        else:
                            rt = random.uniform(30, 60)
                            loggeruser.info(f"发生错误, 等待 {rt:.0f} 秒后重试其他视频: {e}.")
                            await asyncio.sleep(rt)
                        break
                    finally:
                        try:
                            if not await asyncio.shield(asyncio.wait_for(hide_from_resume(obj), 5)):
                                loggeruser.debug(f"未能成功从最近播放中隐藏视频.")
                            else:
                                loggeruser.info(f"已从最近播放中隐藏该视频.")
                        except asyncio.TimeoutError:
                            loggeruser.debug(f"从最近播放中隐藏视频超时.")
            else:
                loggeruser.warning(f"由于没有成功播放视频, 保活失败, 请重新检查配置.")
                return False
        except httpx.HTTPError as e:
            retry += 1
            if retry > retries:
                loggeruser.warning(f"超过最大重试次数, 保活失败: {e}.")
                return False
            else:
                rt = random.uniform(30, 60)
                loggeruser.info(f"连接失败, 等待 {rt:.0f} 秒后重试: {e}.")
                await asyncio.sleep(rt)
        except asyncio.CancelledError:
            raise
        except Exception as e:
            loggeruser.warning(f"发生错误, 保活失败.")
            show_exception(e, regular=False)
            return False


async def watch_continuous(emby: Emby, loggeruser: Logger, stream: bool = False):
    """
    主执行函数 - 持续观看.

    参数:
        emby: Emby 客户端
        loggeruser: 日志器
    """
    while True:
        try:
            async for obj in get_random_media(emby):
                total_ticks = obj.object_dict.get("RunTimeTicks")
                if not total_ticks:
                    if stream:
                        loggeruser.warning('"allow_stream" 无法与 "continuous" 共用, 因此已被忽略.')
                    rt = random.uniform(30, 60)
                    loggeruser.info(f"无法获取视频长度, 等待 {rt:.0f} 秒后重试.")
                    await asyncio.sleep(rt)
                    continue
                total_time = total_ticks / 10000000
                loggeruser.info(f'开始尝试播放 "{truncate_str(obj.name, 10)}" (长度 {total_time:.0f} 秒).')
                try:
                    await play(obj, loggeruser, time=total_time)
                except PlayError as e:
                    rt = random.uniform(30, 60)
                    loggeruser.info(f"发生错误, 等待 {rt:.0f} 秒后重试: {e}.")
                    await asyncio.sleep(rt)
                    continue
                finally:
                    try:
                        if not await asyncio.shield(asyncio.wait_for(hide_from_resume(obj), 2)):
                            loggeruser.debug(f"未能成功从最近播放中隐藏视频.")
                    except asyncio.TimeoutError:
                        loggeruser.debug(f"从最近播放中隐藏视频超时.")
        except httpx.HTTPError as e:
            rt = random.uniform(30, 60)
            loggeruser.info(f"连接失败, 等待 {rt:.0f} 秒后重试: {e}.")
            await asyncio.sleep(rt)
        except asyncio.CancelledError:
            raise
        except Exception as e:
            loggeruser.warning(f"发生错误, 停止持续播放.")
            show_exception(e, regular=False)
            return False


async def watcher(config: dict, instant: bool = False):
    """入口函数 - 观看一个视频."""

    async def wrapper(
        sem: asyncio.Semaphore,
        emby: Emby,
        loggeruser: Logger,
        time: float,
        multiple: bool,
        stream: bool,
    ):
        async with sem:
            try:
                if not instant:
                    wait = random.uniform(180, 360)
                    loggeruser.info(f"播放视频前随机等待 {wait:.0f} 秒.")
                    await asyncio.sleep(wait)
                if isinstance(time, Iterable):
                    tm = max(time) * 4
                else:
                    tm = time * 4
                if multiple:
                    return await asyncio.wait_for(
                        watch_multiple(emby, loggeruser, time, stream), max(tm, 600)
                    )
                else:
                    return await asyncio.wait_for(watch(emby, loggeruser, time, stream), max(tm, 600))
            except asyncio.TimeoutError:
                loggeruser.warning(f"一定时间内未完成播放, 保活失败.")
                return False

    logger.info("开始执行 Emby 保活.")
    tasks = []
    concurrent = int(config.get("watch_concurrent", 3))
    if not concurrent:
        concurrent = 100000
    sem = asyncio.Semaphore(concurrent)
    async for emby, loggeruser, time, multiple, stream in login(config):
        tasks.append(wrapper(sem, emby, loggeruser, time, multiple, stream))
    if not tasks:
        logger.info("没有指定相关的 Emby 服务器, 跳过保活.")
    results = await asyncio.gather(*tasks)
    fails = len(tasks) - sum(results)
    if fails:
        logger.error(f"保活失败 ({fails}/{len(tasks)}).")


async def watcher_schedule(
    config: dict,
    start_time=time(11, 0),
    end_time=time(23, 0),
    days: Union[int, Tuple[int, int]] = 7,
    instant: bool = False,
):
    """计划任务 - 观看一个视频."""

    timestamp_file = Path(config["basedir"]) / "watcher_schedule_next_timestamp"
    current_config = {
        "start_time": start_time.strftime("%H:%M"),
        "end_time": end_time.strftime("%H:%M"),
        "days": days if isinstance(days, int) else list(days),
    }

    while True:
        next_dt = None
        config_changed = False

        if timestamp_file.exists():
            try:
                stored_data = json.loads(timestamp_file.read_text())
                if not isinstance(stored_data, dict):
                    raise ValueError("invalid cache")
                stored_timestamp = stored_data["timestamp"]
                stored_config = stored_data["config"]

                if stored_config != current_config:
                    logger.info("计划任务配置已更改，将重新计算下次执行时间.")
                    config_changed = True
                else:
                    next_dt = datetime.fromtimestamp(stored_timestamp)
                    if next_dt > datetime.now():
                        logger.info(f"从缓存中读取到下次保活时间: {next_dt.strftime('%m-%d %H:%M %p')}.")
            except (ValueError, OSError, json.JSONDecodeError) as e:
                logger.debug(f"读取存储的时间戳失败: {e}")
                config_changed = True

        if not next_dt or next_dt <= datetime.now() or config_changed:
            if isinstance(days, int):
                rand_days = days
            else:
                rand_days = random.randint(*days)
            next_dt = next_random_datetime(start_time, end_time, interval_days=rand_days)
            logger.info(f"下一次保活将在 {next_dt.strftime('%m-%d %H:%M %p')} 进行.")

            try:
                save_data = {"timestamp": next_dt.timestamp(), "config": current_config}
                timestamp_file.write_text(json.dumps(save_data))
            except OSError as e:
                logger.debug(f"存储时间戳失败: {e}")

        await asyncio.sleep((next_dt - datetime.now()).total_seconds())
        try:
            timestamp_file.unlink(missing_ok=True)
        except OSError as e:
            logger.debug(f"删除时间戳文件失败: {e}")
        await watcher(config, instant=instant)


async def watcher_continuous(config: dict):
    """入口函数 - 持续观看."""

    async def wrapper(emby: Emby, loggeruser: Logger, time: float, stream: bool):
        if time:
            if isinstance(time, Iterable):
                time = random.uniform(*time)
            loggeruser.info(f"即将连续播放视频, 持续 {time:.0f} 秒.")
        else:
            loggeruser.info(f"即将无限连续播放视频.")
        try:
            await asyncio.wait_for(watch_continuous(emby, loggeruser, stream), time)
        except asyncio.TimeoutError:
            loggeruser.info(f"连续播放结束, 将在明天继续连续播放.")
            return True
        else:
            return False

    logger.info("开始执行 Emby 持续观看.")
    tasks = []
    async for emby, loggeruser, time, _, stream in login(config, continuous=True):
        tasks.append(wrapper(emby, loggeruser, time, stream))
    if not tasks:
        logger.info("没有指定相关的 Emby 服务器, 跳过持续观看.")
    return await asyncio.gather(*tasks)


async def watcher_continuous_schedule(
    config: dict, start_time=time(11, 0), end_time=time(23, 0), days: int = 1
):
    """计划任务 - 持续观看."""

    timestamp_file = Path(config["basedir"]) / "watcher_continuous_schedule_next_timestamp"
    current_config = {
        "start_time": start_time.strftime("%H:%M"),
        "end_time": end_time.strftime("%H:%M"),
        "days": days,
    }

    while True:
        next_dt = None
        config_changed = False

        if timestamp_file.exists():
            try:
                stored_data = json.loads(timestamp_file.read_text())
                if not isinstance(stored_data, dict):
                    raise ValueError("invalid cache")
                stored_timestamp = stored_data["timestamp"]
                stored_config = stored_data["config"]

                if stored_config != current_config:
                    logger.bind(scheme="embywatcher").info("计划任务配置已更改，将重新计算下次执行时间.")
                    config_changed = True
                else:
                    next_dt = datetime.fromtimestamp(stored_timestamp)
                    if next_dt > datetime.now():
                        logger.bind(scheme="embywatcher").info(
                            f"从缓存中读取到下次持续观看时间: {next_dt.strftime('%m-%d %H:%M %p')}."
                        )
            except (ValueError, OSError, json.JSONDecodeError) as e:
                logger.debug(f"读取存储的时间戳失败: {e}")
                config_changed = True

        if not next_dt or next_dt <= datetime.now() or config_changed:
            next_dt = next_random_datetime(start_time, end_time, interval_days=days)
            logger.bind(scheme="embywatcher").info(
                f"下次持续观看将在 {next_dt.strftime('%m-%d %H:%M %p')} 开始."
            )
            try:
                save_data = {"timestamp": next_dt.timestamp(), "config": current_config}
                timestamp_file.write_text(json.dumps(save_data))
            except OSError as e:
                logger.debug(f"存储时间戳失败: {e}")

        await asyncio.sleep((next_dt - datetime.now()).total_seconds())
        try:
            timestamp_file.unlink(missing_ok=True)
        except OSError as e:
            logger.debug(f"删除时间戳文件失败: {e}")
        t = asyncio.create_task(watcher_continuous(config))
        if not t.done():
            t.cancel()



from ._base import BotCheckin

__ignore__ = True


class XraySGKCheckin(BotCheckin):
    name = "Xray 社工库"
    bot_username = "Zonesgk_bot"
    bot_checkin_cmd = "/qd"
    additional_auth = ["prime"]



from ._base import BotCheckin

from pyrogram.types import Message

__ignore__ = True


class UnionSGKCheckin(BotCheckin):
    name = "银联社工库"
    bot_username = "unionpaysgkbot"
    bot_checkin_cmd = "/start"
    additional_auth = ["prime"]
    bot_checked_keywords = ["今日已签到"]

    async def message_handler(self, client, message: Message):
        if message.text and "欢迎使用" in message.text and message.reply_markup:
            keys = [k.text for r in message.reply_markup.inline_keyboard for k in r]
            for k in keys:
                if "签到" in k:
                    try:
                        await message.click(k)
                    except TimeoutError:
                        pass
                    return
            else:
                self.log.warning(f"签到失败: 账户错误.")
                return await self.fail()
        await super().message_handler(client, message)



from ._base import Monitor

__ignore__ = True


class TestFollowMonitor(Monitor):
    name = "从众 测试"
    chat_name = "api_group"
    chat_allow_outgoing = True
    chat_keyword = r"从众"
    chat_follow_user = 3
    chat_delay = 1
    chat_reply = "我来"



from .pornemby_answer import _PornembyAnswerAnswerMonitor

__ignore__ = True


class TestPornembyExamMonitor(_PornembyAnswerAnswerMonitor):
    name = "Pornemby 科举 测试"
    chat_name = "api_group"
    chat_user = "embykeeper_test_bot"
    chat_allow_outgoing = True



import asyncio
import json
import random
from urllib.parse import urlencode, urlunparse
import uuid
import warnings

import httpx
from loguru import logger

with warnings.catch_warnings():
    warnings.filterwarnings("ignore", category=DeprecationWarning)
    from embypy.emby import Emby as _Emby
    from embypy.objects import EmbyObject
    from embypy.utils.asyncio import async_func
    from embypy.utils.connector import Connector as _Connector


from .. import __version__

logger = logger.bind(scheme="embywatcher")


class Connector(_Connector):
    """重写的 Emby 连接器, 以支持代理."""

    def __init__(
        self,
        url,
        proxy=None,
        ua=None,
        device=None,
        client=None,
        cf_clearance=None,
        **kw,
    ):
        super().__init__(url, **kw)
        self.proxy = proxy
        self.ua = ua
        self.device = device
        self.client = client
        self.fake_headers = self.get_fake_headers()
        self.watch = asyncio.create_task(self.watchdog())
        self.cf_clearance = cf_clearance

    async def watchdog(self, timeout=60):
        logger.debug("Emby 链接池看门狗启动.")
        try:
            counter = {}
            while True:
                await asyncio.sleep(10)
                for s, u in self._session_uses.items():
                    try:
                        if u and u <= 0:
                            if s in counter:
                                counter[s] += 1
                                if counter[s] >= timeout / 10:
                                    logger.debug("销毁了 Emby Session")
                                    async with await self._get_session_lock():
                                        counter[s] = 0
                                        await self._sessions[s].aclose()
                                        self._sessions[s] = None
                                        self._session_uses[s] = None
                            else:
                                counter[s] = 1
                        else:
                            counter.pop(s, None)
                    except (TypeError, KeyError):
                        pass
        except asyncio.CancelledError:
            for s in self._sessions.values():
                if s:
                    try:
                        await asyncio.wait_for(s.aclose(), 1)
                    except asyncio.TimeoutError:
                        pass

    @staticmethod
    def get_device_uuid():
        rd = random.Random()
        rd.seed(uuid.getnode())
        return uuid.UUID(int=rd.getrandbits(128))

    def get_fake_headers(self):
        headers = {}
        ios_uas = [
            "CFNetwork/1335.0.3 Darwin/21.6.0",
            "CFNetwork/1406.0.4 Darwin/22.4.0",
            "CFNetwork/1333.0.4 Darwin/21.5.0",
        ]
        client = "Filebox" if not self.client else self.client
        device = "iPhone" if not self.device else self.device
        device_id = str(self.get_device_uuid()).upper() if not self.device_id else self.device_id
        version = f"1.3.{random.randint(13, 15)}"
        ua = f"Fileball/230 {random.choice(ios_uas)}" if not self.ua else self.ua
        auth_headers = {
            "Client": client,
            "Device": device,
            "DeviceId": device_id,
            "Version": version,
        }
        auth_header = f"Emby {','.join([f'{k}={v}' for k, v in auth_headers.items()])}"
        if self.token:
            headers["X-Emby-Token"] = self.token
        headers["User-Agent"] = ua
        headers["X-Emby-Authorization"] = auth_header
        headers["Accept-Language"] = "zh-CN,zh-Hans;q=0.9"
        headers["Content-Type"] = "application/json"
        headers["Accept"] = "*/*"
        return headers

    async def _get_session(self):
        try:
            loop = asyncio.get_running_loop()
            loop_id = hash(loop)
            async with await self._get_session_lock():
                session = self._sessions.get(loop_id)
                if not session:
                    if self.proxy:
                        proxy = f"{self.proxy['scheme']}://"
                        if self.proxy.get("username"):
                            proxy += f"{self.proxy['username']}:{self.proxy['password']}@"
                        proxy += f"{self.proxy['hostname']}:{self.proxy['port']}"
                    else:
                        proxy = None

                    cookies = {}
                    if self.cf_clearance:
                        cookies["cf_clearance"] = self.cf_clearance

                    timeout = httpx.Timeout(connect=10.0, read=None, write=10.0, pool=10.0)

                    session = httpx.AsyncClient(
                        http2=True,
                        headers=self.fake_headers,
                        cookies=cookies,
                        proxy=proxy,
                        verify=False,
                        follow_redirects=True,
                        timeout=timeout,
                    )
                    self._sessions[loop_id] = session
                    self._session_uses[loop_id] = 1
                    logger.debug("创建了新的 Emby Session.")
                else:
                    self._session_uses[loop_id] += 1
                return session
        except Exception as e:
            logger.error(f"无法创建 Emby Session: {e}")

    async def _end_session(self):
        loop = asyncio.get_running_loop()
        loop_id = hash(loop)
        async with await self._get_session_lock():
            self._session_uses[loop_id] -= 1

    async def _get_session_lock(self):
        loop = asyncio.get_running_loop()
        return self._session_locks.setdefault(loop, asyncio.Lock())

    async def _reset_session(self):
        async with await self._get_session_lock():
            loop = asyncio.get_running_loop()
            loop_id = hash(loop)
            self._sessions[loop_id] = None
            self._session_uses[loop_id] = 0

    @async_func
    async def login_if_needed(self):
        if not self.token:
            return await self.login()

    @async_func
    async def login(self):
        if not self.username or self.attempt_login:
            return

        self.attempt_login = True
        try:
            data = await self.postJson(
                "/Users/AuthenticateByName",
                data={
                    "Username": self.username,
                    "Pw": self.password,
                },
                send_raw=True,
                format="json",
            )

            self.token = data.get("AccessToken", "")
            self.userid = data.get("User", {}).get("Id")
            self.api_key = self.token

            session: httpx.AsyncClient = await self._get_session()
            auth_header = session.headers["X-Emby-Authorization"]
            auth_header += f',Token="{self.token}"'
            session.headers["X-MediaBrowser-Token"] = self.token
            session.headers["Authorization"] = auth_header
            session.headers["X-Emby-Authorization"] = auth_header
            await self._end_session()
        finally:
            self.attempt_login = False

    @async_func
    async def _req(self, method, path, params={}, **query):
        query.pop("format", None)
        await self.login_if_needed()
        for i in range(self.tries):
            url = self.get_url(path, **query)
            try:
                resp = await method(url, **params)
            except httpx.HTTPError as e:
                logger.debug(f'连接 "{url}" 失败, 即将重连: {e.__class__.__name__}: {e}')
            else:
                if self.attempt_login and resp.status_code == 401:
                    raise httpx.HTTPError("用户名密码错误")
                if await self._process_resp(resp):
                    return resp
            await asyncio.sleep(random.random() * i + 0.2)
        raise httpx.HTTPError("无法连接到服务器.")

    @async_func
    async def _process_resp(self, resp):
        if (not resp or resp.status_code == 401) and self.username:
            await self.login()
            return False
        if not resp:
            return False
        if resp.status_code in (502, 503, 504):
            await asyncio.sleep(random.random() * 4 + 0.2)
            return False
        return True

    @staticmethod
    @async_func
    async def resp_to_json(resp: httpx.Response):
        try:
            return json.loads(await resp.aread())
        except json.JSONDecodeError:
            raise httpx.HTTPError(
                'Unexpected JSON output (status: {}): "{}"'.format(
                    resp.status_code,
                    (await resp.aread()).decode(),
                )
            )

    @async_func
    async def get(self, path, **query):
        try:
            session = await self._get_session()
            resp: httpx.Response = await self._req(session.get, path, **query)
            return resp.status_code, (await resp.aread()).decode()
        finally:
            await self._end_session()

    @async_func
    async def delete(self, path, **query):
        try:
            session = await self._get_session()
            resp = await self._req(session.delete, path, **query)
            return resp.status_code
        finally:
            await self._end_session()

    @async_func
    async def _post(self, path, return_json, data, send_raw, **query):
        try:
            session = await self._get_session()
            if send_raw:
                params = {"json": data}
            else:
                params = {"data": json.dumps(data)}
            resp: httpx.Response = await self._req(session.post, path, params=params, **query)
            if return_json:
                return await Connector.resp_to_json(resp)
            else:
                return resp.status_code, (await resp.aread()).decode()
        finally:
            await self._end_session()

    @async_func
    async def get_stream_noreturn(self, path, **query):
        try:
            session = await self._get_session()
            url = self.get_url(path, **query)
            async with session.stream(
                "GET",
                url,
                params={"timeout": httpx.Timeout(connect=10.0, read=None, write=10.0, pool=10.0)},
                **query,
            ) as resp:
                async for _ in resp.aiter_bytes(chunk_size=4096):
                    await asyncio.sleep(random.uniform(5, 10))
        finally:
            await self._end_session()

    def get_url(self, path="/", websocket=False, remote=True, userId=None, pass_uid=False, **query):
        userId = userId or self.userid
        if pass_uid:
            query["userId"] = userId

        if remote:
            url = self.urlremote or self.url
        else:
            url = self.url

        if websocket:
            scheme = url.scheme.replace("http", "ws")
        else:
            scheme = url.scheme

        url = urlunparse((scheme, url.netloc, path, "", "{params}", "")).format(
            UserId=userId, ApiKey=self.api_key, params=urlencode(query)
        )

        return url[:-1] if url[-1] == "?" else url

    @async_func
    async def getJson(self, path, **query):
        try:
            session = await self._get_session()
            resp = await self._req(session.get, path, **query)
            return await Connector.resp_to_json(resp)
        finally:
            await self._end_session()


class Emby(_Emby):
    def __init__(self, url, **kw):
        """重写的 Emby 类, 以支持代理."""
        connector = Connector(url, **kw)
        EmbyObject.__init__(self, {"ItemId": "", "Name": ""}, connector)
        self._partial_cache = {}
        self._cache_lock = asyncio.Condition()

    @async_func
    async def get_items(
        self,
        types,
        path="/Users/{UserId}/Items",
        fields=None,
        limit=10,
        sort="SortName",
        ascending=True,
        **kw,
    ):
        if not fields:
            fields = ["Path", "ParentId", "Overview", "PremiereDate", "DateCreated"]
        resp = await self.connector.getJson(
            path,
            remote=False,
            format="json",
            recursive="true",
            includeItemTypes=",".join(types),
            fields=fields,
            sortBy=sort,
            sortOrder="Ascending" if ascending else "Descending",
            limit=limit,
            **kw,
        )
        return await self.process(resp)



import asyncio
from datetime import date, datetime, time
from pathlib import Path
from typing import List, Tuple

from loguru import logger
from pyrogram.types import User
import yaml

from ...data import get_data
from ...utils import show_exception, truncate_str, distribute_numbers
from ..link import Link
from ..tele import ClientsSession


class SmartMessager:
    """自动智能水群类."""

    name: str = None  # 水群器名称
    chat_name: str = None  # 群聊的名称
    default_messages: str = None  # 语言风格参考话术列表资源名
    additional_auth: List[str] = []  # 额外认证要求
    min_interval: int = None  # 预设两条消息间的最小间隔时间
    max_interval: int = None  # 预设两条消息间的最大间隔时间
    at: Tuple[time, time] = None  # 可发送的时间范围
    msg_per_day: int = 10  # 每天发送的消息数量
    min_msg_gap = 5  # 最小消息间隔

    site_last_message_time = None
    site_lock = asyncio.Lock()

    def __init__(self, account, me: User = None, nofail=True, proxy=None, basedir=None, config: dict = None):
        """
        自动智能水群类.
        参数:
            account: 账号登录信息
            me: 当前用户
            nofail: 启用错误处理外壳, 当错误时报错但不退出
            basedir: 文件存储默认位置
            proxy: 代理配置
            config: 当前水群器的特定配置
        """
        self.account = account
        self.nofail = nofail
        self.proxy = proxy
        self.basedir = basedir
        self.config = config
        self.me = me

        self.min_interval = config.get(
            "min_interval", config.get("interval", self.min_interval or 60)
        )  # 两条消息间的最小间隔时间
        self.max_interval = config.get("max_interval", self.max_interval)  # 两条消息间的最大间隔时间
        self.log = logger.bind(scheme="telemessager", name=self.name, username=me.name)
        self.timeline: List[int] = []  # 消息计划序列
        self.example_messages = []

    async def get_spec_path(self, spec):
        """下载话术文件对应的本地或云端文件."""
        if not Path(spec).exists():
            return await get_data(self.basedir, spec, proxy=self.proxy, caller=f"{self.name}水群")
        else:
            return spec

    async def _start(self):
        """自动水群器的入口函数的错误处理外壳."""
        try:
            return await self.start()
        except asyncio.CancelledError:
            raise
        except Exception as e:
            if self.nofail:
                self.log.warning(f"发生错误, 自动水群器将停止.")
                show_exception(e, regular=False)
                return False
            else:
                raise

    async def start(self):
        """自动水群器的入口函数."""

        async with ClientsSession([self.account], proxy=self.proxy, basedir=self.basedir) as clients:
            async for tg in clients:
                if self.additional_auth:
                    for a in self.additional_auth:
                        if not await Link(tg).auth(a, log_func=self.log.info):
                            return False

            if self.max_interval and self.min_interval > self.max_interval:
                self.log.warning(f"发生错误: 最小间隔不应大于最大间隔, 自动水群将停止.")
                return False

            if not await self.init():
                self.log.warning(f"状态初始化失败, 自动水群将停止.")
                return False

            messages_spec = self.config.get("messages", self.default_messages)
            if messages_spec and (not isinstance(messages_spec, str)):
                self.log.warning(f"发生错误: 参考语言风格列表只能为字符串, 代表远端或本地文件.")
                return False

            if messages_spec:
                messages_file = await self.get_spec_path(messages_spec)
                with open(messages_file, "r") as f:
                    data = yaml.safe_load(f)
                    self.example_messages = data.get("messages", [])[:100]

            self.log.bind(username=tg.me.name).info(
                f"即将预测当前状态下应该发送的水群消息, 但不会实际发送, 仅用于测试."
            )

            await self.send(dummy=True)

        if self.at:
            start_time, end_time = self.at
        else:
            start_time = time(9, 0, 0)
            end_time = time(23, 0, 0)

        start_datetime = datetime.combine(date.today(), start_time)
        end_datetime = datetime.combine(date.today(), end_time)

        start_timestamp = start_datetime.timestamp()
        end_timestamp = end_datetime.timestamp()

        msg_per_day = self.config.get("msg_per_day", self.msg_per_day)

        self.timeline = distribute_numbers(
            start_timestamp, end_timestamp, msg_per_day, self.min_interval, self.max_interval
        )

        # 检查并调整早于当前时间的时间点到明天
        now_timestamp = datetime.now().timestamp()
        for i in range(len(self.timeline)):
            if self.timeline[i] < now_timestamp:
                self.timeline[i] += 86400

        self.timeline = sorted(self.timeline)

        if self.timeline:
            while True:
                dt = datetime.fromtimestamp(self.timeline[0])
                self.log.info(f"下一次发送将在 [blue]{dt.strftime('%m-%d %H:%M:%S')}[/] 进行.")
                sleep_time = max(self.timeline[0] - datetime.now().timestamp(), 0)
                await asyncio.sleep(sleep_time)
                await self.send()
                self.timeline.pop(0)
                if not self.timeline:
                    break

    async def init(self):
        """可重写的初始化函数, 返回 False 将视为初始化错误."""
        return True

    async def send(self, dummy: bool = False):
        async with ClientsSession([self.account], proxy=self.proxy, basedir=self.basedir) as clients:
            async for tg in clients:
                log = self.log.bind(username=tg.me.name)
                chat = await tg.get_chat(self.chat_name)

                context = []
                i = 0
                async for msg in tg.get_chat_history(chat.id, limit=50):
                    i += 1
                    if self.min_msg_gap and msg.outgoing and i < self.min_msg_gap:
                        log.info(f"低于发送消息间隔要求 ({i} < {self.min_msg_gap}), 将不发送消息.")
                        return
                    spec = []
                    text = str(msg.caption or msg.text or "")
                    spec.append(f"消息发送时间为 {msg.date}")
                    if msg.photo:
                        spec.append("包含一张照片")
                    if msg.reply_to_message_id:
                        rmsg = await tg.get_messages(chat.id, msg.reply_to_message_id)
                        spec.append(f"回复了消息: {truncate_str(str(rmsg.caption or rmsg.text or ''), 60)}")
                    spec = " ".join(spec)
                    ctx = truncate_str(text, 180)
                    if msg.from_user.name:
                        ctx = f"{msg.from_user.name}说: {ctx}"
                    if spec:
                        ctx += f" ({spec})"
                    context.append(ctx)

                payload = {}
                if context:
                    payload["context"] = list(reversed(context))
                if self.example_messages:
                    payload["messages"] = self.example_messages

                answer, _ = await Link(tg).infer_msg(payload)

                if answer:
                    if len(answer) > 50:
                        log.info(f"智能推测水群内容过长, 将不发送消息.")
                    elif "SKIP" in answer:
                        log.info(f"智能推测此时不应该水群, 将不发送消息.")
                    else:
                        if dummy:
                            log.info(
                                f'当前情况下在聊天 "{chat.name}" 中推断可发送水群内容为: [gray50]{truncate_str(answer, 20)}[/]'
                            )
                        else:
                            log.info(
                                f'即将在5秒后向聊天 "{chat.name}" 发送: [gray50]{truncate_str(answer, 20)}[/]'
                            )
                            await asyncio.sleep(5)
                            msg = await tg.send_message(chat.id, answer)
                            log.info(f'已向聊天 "{chat.name}" 发送: [gray50]{truncate_str(answer, 20)}[/]')
                            return msg
                else:
                    log.warning(f"智能推测水群内容失败, 将不发送消息.")



from .pornemby_nohp import PornembyNoHPMonitor

__ignore__ = True


class TestPornembyNoHPMonitor(PornembyNoHPMonitor):
    name = "Pornemby 血量耗尽停止发言 测试"
    chat_name = "api_group"
    chat_user = "embykeeper_test_bot"
    chat_allow_outgoing = True



from ._base import BotCheckin

__ignore__ = True


class ShziCheckin(BotCheckin):
    name = "Shzi"
    bot_username = "aishuazibot"
    bot_checkin_cmd = "📅 签到"
    additional_auth = ["prime"]
    max_retries = 6



from ._base import BotCheckin

__ignore__ = True


class DingdangSGKCheckin(BotCheckin):
    name = "叮当猫社工库"
    bot_username = "DingDangCats_Bot"
    bot_checkin_cmd = "/qd"
    additional_auth = ["prime"]



import asyncio
import operator

import aiofiles
import yaml
from dateutil import parser
from loguru import logger
from pyrogram import filters
from pyrogram.enums import ChatType
from pyrogram.handlers import (
    EditedMessageHandler,
    MessageHandler,
    RawUpdateHandler,
)
from pyrogram.types import InlineKeyboardMarkup, Message, ReplyKeyboardMarkup
from rich import box
from rich.live import Live
from rich.panel import Panel
from rich.table import Column, Table
from rich.text import Text

from ..utils import async_partial, batch, flatten, idle, time_in_range
from .tele import Client, ClientsSession

log = logger.bind(scheme="debugtool")


async def _dump_message(client: Client, message: Message, table: Table):
    """消息调试工具, 将消息更新列到 table 中."""
    text = message.text or message.caption
    if text:
        text = text.replace("\n", " ")
        if not text:
            return
    else:
        return
    if message.from_user:
        user = message.from_user
        sender_id = str(user.id)
        sender_icon = "👤"
        if message.outgoing:
            sender = Text("Me", style="bold red")
            text = Text(text, style="red")
        else:
            sender = user.name
            if user.is_bot:
                sender_icon = "🤖"
                sender = Text(sender, style="bold yellow")
    else:
        sender = sender_id = sender_icon = None

    chat_id = "{: }".format(message.chat.id)
    if message.chat.type == ChatType.GROUP or message.chat.type == ChatType.SUPERGROUP:
        chat = message.chat.title
        chat_icon = "👥"
    elif message.chat.type == ChatType.CHANNEL:
        chat = message.chat.title
        chat_icon = "📢"
    elif message.chat.type == ChatType.BOT:
        chat = None
        chat_icon = "🤖"
    else:
        chat = chat_icon = None
    others = []
    if message.photo:
        others.append(f"照片: {message.photo.file_unique_id}")
    if message.reply_markup:
        if isinstance(message.reply_markup, InlineKeyboardMarkup):
            key_info = "|".join([k.text for r in message.reply_markup.inline_keyboard for k in r])
            others.append(f"按钮: {key_info}")
        elif isinstance(message.reply_markup, ReplyKeyboardMarkup):
            key_info = "|".join([k.text for r in message.reply_markup.keyboard for k in r])
            others.append(f"按钮: {key_info}")
    return table.add_row(
        f"{client.me.name}",
        "│",
        chat_icon,
        chat,
        chat_id,
        "│",
        sender_icon,
        sender,
        sender_id,
        "│",
        text,
        "|",
        "; ".join(others),
    )


async def follower(config: dict):
    """消息调试工具入口函数."""
    columns = [
        Column("用户", style="cyan", justify="center"),
        Column("", max_width=1, style="white"),
        Column("", max_width=2, overflow="crop"),
        Column("会话", style="bright_blue", no_wrap=True, justify="right", max_width=15),
        Column("(ChatID)", style="gray50", no_wrap=True, max_width=20),
        Column("", max_width=1, style="white"),
        Column("", max_width=2, overflow="crop"),
        Column("发信人", style="green", no_wrap=True, max_width=15, justify="right"),
        Column("(UserID)", style="gray50", no_wrap=True, max_width=15),
        Column("", max_width=1, style="white"),
        Column("信息", no_wrap=False, min_width=30, max_width=50),
        Column("", max_width=1, style="white"),
        Column("其他", no_wrap=False, min_width=30, max_width=50),
    ]
    async with ClientsSession.from_config(config) as clients:
        table = Table(*columns, header_style="bold magenta", box=box.SIMPLE)
        func = async_partial(_dump_message, table=table)
        async for tg in clients:
            await tg.add_handler(MessageHandler(func))
            await tg.add_handler(EditedMessageHandler(func))
        with Live(table, refresh_per_second=4, vertical_overflow="visible"):
            await idle()


def _dumper_mask_update(update):
    if getattr(update, "outgoing", False):
        from_user = getattr(update, "from_user", {})
        if from_user:
            if hasattr(from_user, "id"):
                setattr(from_user, "id", "(sender)")
            if hasattr(from_user, "first_name"):
                setattr(from_user, "first_name", "(sender)")
            if hasattr(from_user, "last_name"):
                setattr(from_user, "last_name", "(sender)")
            if hasattr(from_user, "next_offline_date"):
                setattr(from_user, "next_offline_date", "(sender)")
            if hasattr(from_user, "username"):
                setattr(from_user, "username", "(sender)")
    return update


async def _dumper_raw(client, update, users, chats):
    await client.queue.put(_dumper_mask_update(update))


async def _dumper_update(client, update):
    if isinstance(update, list):
        for u in update:
            await client.queue.put(_dumper_mask_update(u))
    else:
        await client.queue.put(_dumper_mask_update(update))


async def dumper(config: dict, specs=["message"]):
    type_handler = {
        "message": MessageHandler(_dumper_update),
        "edited_message": EditedMessageHandler(_dumper_update),
        "raw": RawUpdateHandler(_dumper_raw),
    }
    queue = asyncio.Queue()
    async with ClientsSession.from_config(config) as clients:
        async for tg in clients:
            tg.queue = queue
            for s in specs:
                try:
                    t, c = s.split("@")
                    c = [i.strip() for i in c.split(",")]
                except ValueError:
                    t = s
                    c = []
                if t == "all":
                    handlers = [type_handler[t] for t in ["message", "edited_message"]]
                    for h in handlers:
                        h.filters = filters.chat(c) if c else None
                        await tg.add_handler(h)
                    break
                else:
                    try:
                        handler = type_handler[t]
                    except KeyError:
                        log.warning(f'更新类型 {t} 不可用, 请选择: {", ".join(list(type_handler.keys()))}')
                        continue
                handler.filters = filters.chat(c) if c else None
                await tg.add_handler(handler)
            log.info(f'开始监控账号: "{tg.me.name}" 中的更新.')
        while True:
            update = str(await queue.get())

            print(update, flush=True)


async def _saver_raw(client, update, users, chats):
    await client.saver_queue.put(update)


async def _saver_dumper(queue, output):
    async with aiofiles.open(output, "w+", buffering=1, encoding="utf-8") as f:
        while True:
            update = await queue.get()
            await f.write(str(update) + "\n")


async def saver(config: dict):
    async with ClientsSession.from_config(config) as clients:
        tasks = []
        async for tg in clients:
            tg.saver_queue = queue = asyncio.Queue()
            output = f"{tg.me.phone_number}.updates.json"
            await tg.add_handler(RawUpdateHandler(_saver_raw), group=10000)
            tasks.append(_saver_dumper(queue, output))
        await asyncio.gather(*tasks)


class IndentDumper(yaml.Dumper):
    """输出带缩进的 YAML."""

    def increase_indent(self, flow=False, indentless=False):
        return super().increase_indent(flow, False)


async def analyzer(config: dict, chats, keywords, timerange, limit=10000, outputs=1000):
    """历史消息分析工具入口函数."""

    from rich.progress import MofNCompleteColumn, Progress, SpinnerColumn

    def render_page(progress, texts):
        page = Table.grid()
        page.add_row(Panel(progress))
        if texts:
            msgs = sorted(texts.items(), key=operator.itemgetter(1), reverse=True)
            columns = flatten([[Column(max_width=15, no_wrap=True), Column(min_width=2)] for _ in range(4)])
            table = Table(*columns, show_header=False, box=box.SIMPLE)
            cols = []
            for col in batch(msgs, 12):
                col = [(t.split()[0], str(c)) for t, c in col]
                col += [("", "")] * (12 - len(col))
                cols.append(col)
                if len(cols) >= 4:
                    break
            for row in map(list, zip(*cols)):
                table.add_row(*flatten(row))
            page.add_row(table)
        return page

    texts = {}
    if timerange:
        start, end = (parser.parse(t).time() for t in timerange)
    async with ClientsSession.from_config(config) as clients:
        async for tg in clients:
            target = f"{tg.me.phone_number}.msgs.yaml"
            log.info(f'开始分析账号: "{tg.me.name}", 结果将写入"{target}".')
            pcs = list(Progress.get_default_columns())
            pcs.insert(0, SpinnerColumn())
            pcs.insert(3, MofNCompleteColumn(table_column=Column(justify="center")))
            p = Progress(*pcs, transient=True)
            with Live(render_page(p, texts)) as live:
                updates = 0
                pchats = p.add_task("[red]会话: ", total=len(chats))
                for c in chats:
                    c = c.rsplit("/", 1)[-1]
                    pmsgs = p.add_task("[red]记录: ", total=limit)
                    m: Message
                    async for m in tg.get_chat_history(c, limit=limit):
                        if m.text:
                            if m.from_user and not m.from_user.is_bot:
                                if (not keywords) or any(s in m.text for s in keywords):
                                    if (not timerange) or time_in_range(start, end, m.date.time()):
                                        if m.text in texts:
                                            texts[m.text] += 1
                                        else:
                                            texts[m.text] = 1
                                        updates += 1
                                        if updates % 200 == 0:
                                            live.update(render_page(p, texts))
                        p.advance(pmsgs)
                    p.update(pmsgs, visible=False)
                    p.advance(pchats)
            with open(target, "w+", encoding="utf-8") as f:
                yaml.dump(
                    {
                        "messages": [
                            str(t) for t, _ in sorted(texts.items(), key=operator.itemgetter(1), reverse=True)
                        ][:outputs]
                    },
                    f,
                    default_flow_style=False,
                    encoding="utf-8",
                    allow_unicode=True,
                    Dumper=IndentDumper,
                )



import asyncio
import random
import string

from thefuzz import process
from pyrogram.types import Message

from ..link import Link
from ._base import BotCheckin


class CCCheckin(BotCheckin):
    name = "CC公益"
    bot_username = "EmbyCc_bot"
    bot_checkin_cmd = "/start"
    bot_checked_keywords = ["已经签到过了"]
    bot_checkin_caption_pat = "请选择正确验证码"
    max_retries = 1
    additional_auth = ["ocr"]

    async def message_handler(self, client, message: Message):
        if message.caption and "欢迎使用" in message.caption and message.reply_markup:
            keys = [k.text for r in message.reply_markup.inline_keyboard for k in r]
            for k in keys:
                if "签到" in k:
                    try:
                        await message.click(k)
                    except TimeoutError:
                        pass
                    return
            else:
                self.log.warning(f"签到失败: 账户错误.")
                return await self.fail()
        await super().message_handler(client, message)

    async def on_photo(self, message: Message):
        """分析分析传入的验证码图片并返回验证码."""
        if not message.reply_markup:
            return
        for i in range(3):
            result: str = await Link(self.client).ocr(message.photo.file_id)
            if result:
                self.log.debug(f"远端已解析答案: {result}.")
                break
            else:
                self.log.warning(f"远端解析失败, 正在重试解析 ({i + 1}/3).")
        else:
            self.log.warning(f"签到失败: 验证码识别错误.")
            return await self.fail()
        options = [k.text for r in message.reply_markup.inline_keyboard for k in r]
        result = result.translate(str.maketrans("", "", string.punctuation)).replace(" ", "")
        captcha, score = process.extractOne(result, options)
        if score < 50:
            self.log.warning(f"远端答案难以与可用选项相匹配 (分数: {score}/100).")
        self.log.debug(f"[gray50]接收验证码: {captcha}.[/]")
        await asyncio.sleep(random.uniform(2, 4))
        try:
            await message.click(captcha)
        except TimeoutError:
            pass



from ._base import BotCheckin


class PandaTVGroupCheckin(BotCheckin):
    name = "PandaTV 群组发言"
    chat_name = "PandaTV_Emby_Bot"
    bot_checkin_cmd = "签到"
    skip = 14

    async def send_checkin(self, retry=False):
        await super().send_checkin(retry=retry)
        self.finished.set()



import asyncio
import random
from urllib.parse import parse_qs, urlencode, urlparse, urljoin

from pyrogram import filters
from pyrogram.types import Message
from pyrogram.raw.functions.messages import RequestWebView
from aiohttp import ClientSession, TCPConnector
from aiohttp_socks import ProxyConnector, ProxyTimeoutError, ProxyError, ProxyType
from faker import Faker

from ..link import Link
from ._base import Monitor


class FutureMonitor(Monitor):
    name = "未响"
    chat_keyword = r"FutureEcho-register-\w{8}-\w{4}-\w{3}-\w{5}"
    bot_username = "lotayu_bot"
    chat_name = "FutureEcho_Chat"
    notify_create_name = True
    allow_edit = False
    additional_auth = ["captcha"]

    async def solve_captcha(self, url: str):
        token = await Link(self.client).captcha("future_echo")
        if not token:
            return False
        else:
            scheme = urlparse(url)
            params = parse_qs(scheme.query)
            url_submit = scheme._replace(path="/x/api/submit", query="", fragment="").geturl()
            uuid = params.get("id", [None])[0]
            if self.proxy:
                connector = ProxyConnector(
                    proxy_type=ProxyType[self.proxy["scheme"].upper()],
                    host=self.proxy["hostname"],
                    port=self.proxy["port"],
                    username=self.proxy.get("username", None),
                    password=self.proxy.get("password", None),
                )
            else:
                connector = TCPConnector()
            origin = scheme._replace(path="/", query="", fragment="").geturl()
            useragent = Faker().safari()
            headers = {
                "Content-Type": "application/x-www-form-urlencoded",
                "Referer": url,
                "Origin": origin,
                "User-Agent": useragent,
            }
            data = {
                "uuid": uuid,
                "cf-turnstile-response": token,
            }
            try:
                async with ClientSession(connector=connector) as session:
                    async with session.post(url_submit, headers=headers, data=data) as resp:
                        result = await resp.text()
                        if "完成" in result:
                            return True
            except:
                return False

    async def on_trigger(self, message: Message, key, reply):
        for i in range(3):
            if i:
                self.log.info(f"正在重试注册 ({i}/3).")
            msg = await self.client.wait_reply(self.bot_username, f"/start")
            text = msg.text or msg.caption
            if "你还未加入" in text:
                self.log.error("账户错误, 无法注册.")
                return
            async with self.client.catch_reply(self.bot_username) as f1:
                async with self.client.catch_edit(msg, ~filters.regex("请先完成验证")) as f2:
                    try:
                        msg = await msg.click("💡註冊帳戶", timeout=1)
                    except TimeoutError:
                        pass
                    except ValueError:
                        self.log.error("未能找到注册按钮, 无法注册.")
                        return
                    try:
                        done, pending = await asyncio.wait(
                            [f1, f2], return_when=asyncio.FIRST_COMPLETED, timeout=10
                        )
                    except asyncio.TimeoutError:
                        self.log.error("点击注册按钮无响应, 无法注册.")
                        return
                    else:
                        for f in pending:
                            f.cancel()
                        msg = list(done)[0].result()
            text = msg.text or msg.caption
            if "验证您的身份" in text:
                self.log.info("需要验证身份, 正在解析.")
                url = None
                if msg.reply_markup:
                    buttons = [button for line in msg.reply_markup.inline_keyboard for button in line]
                    for b in buttons:
                        if "Verify" in b.text and b.web_app:
                            url = b.web_app.url
                            break
                if not url:
                    self.log.error("需要验证身份但没有找到 URL, 无法注册.")
                    return
                bot_peer = await self.client.resolve_peer(self.bot_username)
                url_auth = (
                    await self.client.invoke(
                        RequestWebView(peer=bot_peer, bot=bot_peer, platform="ios", url=url)
                    )
                ).url
                if not await self.solve_captcha(url_auth):
                    self.log.error("验证码解析失败, 无法注册.")
                    return
                else:
                    await asyncio.sleep(random.uniform(3, 5))
                    self.log.info("已成功验证, 重新进行注册流程.")
                    continue
            else:
                if ("邀請碼" in text) or ("邀请码" in text):
                    msg = await self.client.wait_reply(self.bot_username, key)
                    text = msg.text or msg.caption
                    if "无效" in text:
                        self.log.error("邀请码无效, 无法注册.")
                        return
                if "用户名" in text:
                    msg = await self.client.wait_reply(self.bot_username, self.unique_name)
                    text = msg.text or msg.caption
                if "邮箱地址" in text:
                    msg = await self.client.wait_reply(self.bot_username, f"{self.unique_name}@gmail.com")
                    text = msg.text or msg.caption
                if "创建成功" in text:
                    self.log.bind(msg=True).info(f"已在 Bot @{self.bot_username} 成功创建用户, 请查看.")
                    return
                else:
                    self.log.warning(f"注册失败, 可能是注册流程出错, 请检查.")
                    return



from ._base import BotCheckin

__ignore__ = True


class ChunjiangSGKCheckin(BotCheckin):
    name = "春江社工库"
    bot_username = "ChunJiang_SGK_Bot"
    bot_checkin_cmd = "/qd"
    additional_auth = ["prime"]



from __future__ import annotations

import base64
import binascii
from collections import OrderedDict
from contextlib import asynccontextmanager
from datetime import datetime, timezone
import asyncio
import getpass
import inspect
from pathlib import Path
import pickle
import random
import struct
import sys
from typing import AsyncGenerator, Optional, Union
from sqlite3 import OperationalError
import logging
import tempfile
import typing

from rich.prompt import Prompt
from appdirs import user_data_dir
from loguru import logger
import pyrogram
from pyrogram import raw, types, utils, filters, dispatcher, session
from pyrogram.enums import SentCodeType
from pyrogram.errors import (
    ChannelPrivate,
    PersistentTimestampOutdated,
    PersistentTimestampInvalid,
    BadRequest,
    RPCError,
    ApiIdPublishedFlood,
    Unauthorized,
    SessionPasswordNeeded,
    CodeInvalid,
    PhoneCodeInvalid,
    BadMsgNotification,
    FloodWait,
    PhoneNumberInvalid,
    PhoneNumberBanned,
)
from pyrogram.storage.storage import Storage
from pyrogram.handlers import (
    MessageHandler,
    RawUpdateHandler,
    DisconnectHandler,
    EditedMessageHandler,
)

from pyrogram.handlers.handler import Handler
from aiocache import Cache
import aiohttp
from aiohttp_socks import ProxyConnector, ProxyType, ProxyConnectionError, ProxyTimeoutError

from embykeeper import var, __name__ as __product__, __version__
from embykeeper.utils import async_partial, show_exception, to_iterable

if typing.TYPE_CHECKING:
    from telethon import TelegramClient

logger = logger.bind(scheme="telegram")

_id = b"\x80\x04\x95\x15\x00\x00\x00\x00\x00\x00\x00]\x94(K2K3K7K8K5K8K4K6e."
_hash = b"\x80\x04\x95E\x00\x00\x00\x00\x00\x00\x00]\x94(KbKdK7K4K0KeKaK2KaKcKeKeK7K3K9K0KeK0KbK3K5K4KeKcK8K0K9KcK8K7K0Kfe."
_decode = lambda x: "".join(map(chr, to_iterable(pickle.loads(x))))

API_KEY = {
    "_": {"api_id": _decode(_id), "api_hash": _decode(_hash)}
    # "nicegram": {"api_id": "94575", "api_hash": "a3406de8d171bb422bb6ddf3bbd800e2"},
    # "tgx-android": {"api_id": "21724", "api_hash": "3e0cb5efcd52300aec5994fdfc5bdc16"},
    # "tg-react": {"api_id": "414121", "api_hash": "db09ccfc2a65e1b14a937be15bdb5d4b"},
}


def _name(self: Union[types.User, types.Chat]):
    return " ".join([n for n in (self.first_name, self.last_name) if n])


def _chat_name(self: types.Chat):
    if self.title:
        return self.title
    else:
        return _name(self)


setattr(types.User, "name", property(_name))
setattr(types.Chat, "name", property(_chat_name))


class LogRedirector(logging.StreamHandler):
    def emit(self, record):
        try:
            if record.levelno >= logging.WARNING:
                logger.debug(f"Pyrogram log: {record.getMessage()}")
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            self.handleError(record)


pyrogram_session_logger = logging.getLogger("pyrogram")
for h in pyrogram_session_logger.handlers[:]:
    pyrogram_session_logger.removeHandler(h)
pyrogram_session_logger.addHandler(LogRedirector())


class Dispatcher(dispatcher.Dispatcher):
    def __init__(self, client: Client):
        super().__init__(client)
        self.mutex = asyncio.Lock()

    async def start(self):
        logger.debug("Telegram 更新分配器启动.")
        if not self.client.no_updates:
            self.handler_worker_tasks = [
                self.loop.create_task(self.handler_worker()) for _ in range(self.client.workers)
            ]
            if not self.client.skip_updates:
                await self.client.recover_gaps()

    def add_handler(self, handler, group: int):
        async def fn():
            async with self.mutex:
                if group not in self.groups:
                    self.groups[group] = []
                    self.groups = OrderedDict(sorted(self.groups.items()))
                self.groups[group].append(handler)
                # logger.debug(f"增加了 Telegram 更新处理器: {handler.__class__.__name__}.")

        return self.loop.create_task(fn())

    def remove_handler(self, handler, group: int):
        async def fn():
            async with self.mutex:
                if group not in self.groups:
                    raise ValueError(f"Group {group} does not exist. Handler was not removed.")
                self.groups[group].remove(handler)
                # logger.debug(f"移除了 Telegram 更新处理器: {handler.__class__.__name__}.")

        return self.loop.create_task(fn())

    async def handler_worker(self):
        while True:
            packet = await self.updates_queue.get()

            if packet is None:
                break

            try:
                update, users, chats = packet
                parser = self.update_parsers.get(type(update), None)

                try:
                    parsed_update, handler_type = (
                        await parser(update, users, chats) if parser is not None else (None, type(None))
                    )
                except ValueError:
                    continue

                async with self.mutex:
                    groups = {i: g[:] for i, g in self.groups.items()}

                for group in groups.values():
                    for handler in group:
                        args = None

                        if isinstance(handler, handler_type):
                            try:
                                if await handler.check(self.client, parsed_update):
                                    args = (parsed_update,)
                            except Exception as e:
                                logger.warning(f"Telegram 错误: {e}")
                                continue

                        elif isinstance(handler, RawUpdateHandler):
                            try:
                                if await handler.check(self.client, update):
                                    args = (update, users, chats)
                            except Exception as e:
                                logger.debug(f"更新回调函数内发生错误.")
                                show_exception(e, regular=False)
                        if args is None:
                            continue

                        try:
                            if inspect.iscoroutinefunction(handler.callback):
                                await handler.callback(self.client, *args)
                            else:
                                await self.loop.run_in_executor(
                                    self.client.executor, handler.callback, self.client, *args
                                )
                        except pyrogram.StopPropagation:
                            raise
                        except pyrogram.ContinuePropagation:
                            continue
                        except Exception as e:
                            logger.error(f"更新回调函数内发生错误.")
                            show_exception(e, regular=False)
                        break
                    else:
                        continue
                    break
            except pyrogram.StopPropagation:
                pass
            except TimeoutError:
                logger.info("网络不稳定, 可能遗漏消息.")
            except Exception as e:
                logger.error("更新控制器错误.")
                show_exception(e, regular=False)


class Client(pyrogram.Client):
    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        self.cache = Cache()
        self.lock = asyncio.Lock()
        self.dispatcher = Dispatcher(self)
        self._last_special_invoke = {}
        self._special_invoke_lock = asyncio.Lock()
        self._last_invoke = {}
        self._invoke_lock = asyncio.Lock()
        self._login_time: datetime = None
        self._config_index: int = None

    async def authorize(self):
        if self.bot_token:
            return await self.sign_in_bot(self.bot_token)
        retry = False
        while True:
            try:
                sent_code = await self.send_code(self.phone_number)
                code_target = {
                    SentCodeType.APP: " Telegram 客户端",
                    SentCodeType.SMS: "短信",
                    SentCodeType.CALL: "来电",
                    SentCodeType.FLASH_CALL: "闪存呼叫",
                    SentCodeType.FRAGMENT_SMS: " Fragment 短信",
                    SentCodeType.EMAIL_CODE: "邮件",
                }
                if not self.phone_code:
                    if retry:
                        msg = f'验证码错误, 请重新输入 "{self.phone_number}" 的登录验证码 (按回车确认)'
                    else:
                        msg = f'请从{code_target[sent_code.type]}接收 "{self.phone_number}" 的登录验证码 (按回车确认)'
                    try:
                        self.phone_code = Prompt.ask(" " * 23 + msg, console=var.console)
                    except EOFError:
                        raise BadRequest(
                            f'登录 "{self.phone_number}" 时出现异常: 您正在使用非交互式终端, 无法输入验证码.'
                        )
                signed_in = await self.sign_in(self.phone_number, sent_code.phone_code_hash, self.phone_code)
            except (CodeInvalid, PhoneCodeInvalid):
                self.phone_code = None
                retry = True
            except SessionPasswordNeeded:
                retry = False
                while True:
                    if not self.password:
                        if retry:
                            msg = f'密码错误, 请重新输入 "{self.phone_number}" 的两步验证密码 (不显示, 按回车确认)'
                        else:
                            msg = f'需要输入 "{self.phone_number}" 的两步验证密码 (不显示, 按回车确认)'
                        self.password = Prompt.ask(" " * 23 + msg, password=True, console=var.console)
                    try:
                        return await self.check_password(self.password)
                    except BadRequest:
                        self.password = None
                        retry = True
            except FloodWait:
                raise BadRequest(f'登录 "{self.phone_number}" 时出现异常: 登录过于频繁.')
            except PhoneNumberInvalid:
                raise BadRequest(
                    f'登录 "{self.phone_number}" 时出现异常: 您使用了错误的手机号 (格式错误或没有注册).'
                )
            except PhoneNumberBanned:
                raise BadRequest(f'登录 "{self.phone_number}" 时出现异常: 您的账户已被封禁.')
            except Exception as e:
                logger.error(f"登录时出现异常错误!")
                show_exception(e, regular=False)
                retry = True
            else:
                break
        if isinstance(signed_in, types.User):
            return signed_in
        else:
            raise BadRequest("该账户尚未注册")

    def add_handler(self, handler: Handler, group: int = 0):
        if isinstance(handler, DisconnectHandler):
            self.disconnect_handler = handler.callback

            async def dummy():
                pass

            return asyncio.ensure_future(dummy())
        else:
            return self.dispatcher.add_handler(handler, group)

    def remove_handler(self, handler: Handler, group: int = 0):
        if isinstance(handler, DisconnectHandler):
            self.disconnect_handler = None

            async def dummy():
                pass

            return asyncio.ensure_future(dummy())
        else:
            return self.dispatcher.remove_handler(handler, group)

    async def get_dialogs(
        self, limit: int = 0, exclude_pinned=None, folder_id=None
    ) -> Optional[AsyncGenerator["types.Dialog", None]]:
        async with self.lock:
            cache_id = f"dialogs_{self.phone_number}_{folder_id}_{1 if exclude_pinned else 0}"
            (offset_id, offset_date, offset_peer), cache = await self.cache.get(
                cache_id, ((0, 0, raw.types.InputPeerEmpty()), [])
            )

            current = 0
            total = limit or (1 << 31) - 1
            limit = min(100, total)

            for c in cache:
                yield c
                current += 1
                if current >= total:
                    return

            while True:
                r = await self.invoke(
                    raw.functions.messages.GetDialogs(
                        offset_date=offset_date,
                        offset_id=offset_id,
                        offset_peer=offset_peer,
                        limit=limit,
                        hash=0,
                        exclude_pinned=exclude_pinned,
                        folder_id=folder_id,
                    ),
                    sleep_threshold=60,
                )

                users = {i.id: i for i in r.users}
                chats = {i.id: i for i in r.chats}

                messages = {}

                for message in r.messages:
                    if isinstance(message, raw.types.MessageEmpty):
                        continue

                    chat_id = utils.get_peer_id(message.peer_id)
                    try:
                        messages[chat_id] = await types.Message._parse(self, message, users, chats)
                    except RPCError:
                        continue

                dialogs = []

                for dialog in r.dialogs:
                    if not isinstance(dialog, raw.types.Dialog):
                        continue
                    try:
                        dialogs.append(types.Dialog._parse(self, dialog, messages, users, chats))
                    except BadRequest:
                        continue

                if not dialogs:
                    return

                last = dialogs[-1]

                offset_id = last.top_message.id
                offset_date = utils.datetime_to_timestamp(last.top_message.date)
                offset_peer = await self.resolve_peer(last.chat.id)

                _, cache = await self.cache.get(cache_id, ((0, 0, raw.types.InputPeerEmpty()), []))
                await self.cache.set(
                    cache_id, ((offset_id, offset_date, offset_peer), cache + dialogs), ttl=120
                )
                for dialog in dialogs:
                    yield dialog

                    current += 1

                    if current >= total:
                        return

    @asynccontextmanager
    async def catch_reply(self, chat_id: Union[int, str], outgoing=False, filter=None):
        async def handler_func(client, message, future: asyncio.Future):
            try:
                future.set_result(message)
            except asyncio.InvalidStateError:
                pass

        future = asyncio.Future()
        f = filters.chat(chat_id)
        if not outgoing:
            f = f & (~filters.outgoing)
        if filter:
            f = f & filter
        handler = MessageHandler(async_partial(handler_func, future=future), f)
        await self.add_handler(handler, group=0)
        try:
            yield future
        finally:
            await self.remove_handler(handler, group=0)

    async def invoke(
        self: "pyrogram.Client",
        query: raw.core.TLObject,
        retries: int = session.Session.MAX_RETRIES,
        timeout: float = session.Session.WAIT_TIMEOUT,
        sleep_threshold: float = None,
        business_connection_id: str = None,
    ):
        special_methods = {"SendMessage", "DeleteMessages"}
        except_methods = {"GetMessages", "GetChannelDifference", "GetStickerSet"}

        query_name = query.__class__.__name__

        if query_name in special_methods:
            if self._login_time:
                time_since_login = (datetime.now() - self._login_time).total_seconds()
                if time_since_login < 5:
                    wait_time = 5 - time_since_login
                    logger.info(
                        f"距离登录时间 {time_since_login:.1f} 秒, 等待 {wait_time:.1f} 秒以执行下一步操作."
                    )
                    await asyncio.sleep(wait_time)

            async with self._special_invoke_lock:
                now = datetime.now().timestamp()
                last_invoke = self._last_special_invoke.get(query_name, 0)
                if now - last_invoke < 3:
                    wait_time = 3 - (now - last_invoke)
                    await asyncio.sleep(wait_time)
                self._last_special_invoke[query_name] = datetime.now().timestamp()

        if query_name not in except_methods:
            async with self._invoke_lock:
                now = datetime.now().timestamp()
                last_invoke = self._last_invoke.get(query_name, 0)
                if now - last_invoke < 1:
                    wait_time = 1 - (now - last_invoke)
                    await asyncio.sleep(wait_time)
                self._last_invoke[query_name] = datetime.now().timestamp()

        logger.trace(f"请求: {query_name}")
        return await super().invoke(
            query,
            retries=retries,
            timeout=timeout,
            sleep_threshold=sleep_threshold,
            business_connection_id=business_connection_id,
        )

    @asynccontextmanager
    async def catch_edit(self, message: types.Message, filter=None):
        def filter_message(id: int):
            async def func(flt, _, message: types.Message):
                return message.id == id

            return filters.create(func, "MessageFilter")

        async def handler_func(client, message, future: asyncio.Future):
            try:
                future.set_result(message)
            except asyncio.InvalidStateError:
                pass

        future = asyncio.Future()
        f = filter_message(message.id)
        if filter:
            f = f & filter
        handler = EditedMessageHandler(async_partial(handler_func, future=future), f)
        await self.add_handler(handler, group=0)
        try:
            yield future
        finally:
            await self.remove_handler(handler, group=0)

    async def wait_reply(
        self,
        chat_id: Union[int, str],
        send: str = None,
        timeout: float = 10,
        outgoing=False,
        filter=None,
    ):
        async with self.catch_reply(chat_id=chat_id, filter=filter) as f:
            if send:
                await self.send_message(chat_id, send)
            msg: types.Message = await asyncio.wait_for(f, timeout)
            return msg

    async def wait_edit(
        self,
        message: types.Message,
        click: Union[str, int] = None,
        timeout: float = 10,
        noanswer=True,
        filter=None,
    ):
        async with self.catch_edit(message, filter=filter) as f:
            if click:
                try:
                    await message.click(click)
                except TimeoutError:
                    if noanswer:
                        pass
                    else:
                        raise
            msg: types.Message = await asyncio.wait_for(f, timeout)
            return msg

    async def mute_chat(self, chat_id: Union[int, str], until: Union[int, datetime, None] = None):
        if until is None:
            until = 0x7FFFFFFF  # permanent mute
        elif isinstance(until, datetime):
            until = until.timestamp()
        return await self.invoke(
            raw.functions.account.UpdateNotifySettings(
                peer=raw.types.InputNotifyPeer(peer=await self.resolve_peer(chat_id)),
                settings=raw.types.InputPeerNotifySettings(
                    show_previews=False,
                    mute_until=int(until),
                ),
            )
        )

    async def handle_updates(self, updates):
        if isinstance(updates, (raw.types.Updates, raw.types.UpdatesCombined)):
            is_min = any(
                (
                    await self.fetch_peers(updates.users),
                    await self.fetch_peers(updates.chats),
                )
            )

            users = {u.id: u for u in updates.users}
            chats = {c.id: c for c in updates.chats}

            for update in updates.updates:
                channel_id = getattr(
                    getattr(getattr(update, "message", None), "peer_id", None), "channel_id", None
                ) or getattr(update, "channel_id", None)

                pts = getattr(update, "pts", None)
                pts_count = getattr(update, "pts_count", None)

                if pts and not self.skip_updates:
                    await self.storage.update_state(
                        (
                            utils.get_channel_id(channel_id) if channel_id else 0,
                            pts,
                            None,
                            updates.date,
                            updates.seq,
                        )
                    )

                if isinstance(update, raw.types.UpdateNewChannelMessage) and is_min:
                    message = update.message

                    if not isinstance(message, raw.types.MessageEmpty):
                        try:
                            diff = await self.invoke(
                                raw.functions.updates.GetChannelDifference(
                                    channel=await self.resolve_peer(utils.get_channel_id(channel_id)),
                                    filter=raw.types.ChannelMessagesFilter(
                                        ranges=[
                                            raw.types.MessageRange(
                                                min_id=update.message.id, max_id=update.message.id
                                            )
                                        ]
                                    ),
                                    pts=pts - pts_count,
                                    limit=pts,
                                    force=False,
                                )
                            )
                        except (ChannelPrivate, PersistentTimestampOutdated, PersistentTimestampInvalid):
                            pass
                        else:
                            if not isinstance(diff, raw.types.updates.ChannelDifferenceEmpty):
                                users.update({u.id: u for u in diff.users})
                                chats.update({c.id: c for c in diff.chats})

                self.dispatcher.updates_queue.put_nowait((update, users, chats))
        elif isinstance(updates, (raw.types.UpdateShortMessage, raw.types.UpdateShortChatMessage)):
            if not self.skip_updates:
                await self.storage.update_state((0, updates.pts, None, updates.date, None))

            diff = await self.invoke(
                raw.functions.updates.GetDifference(
                    pts=updates.pts - updates.pts_count, date=updates.date, qts=-1
                )
            )

            if diff.new_messages:
                self.dispatcher.updates_queue.put_nowait(
                    (
                        raw.types.UpdateNewMessage(
                            message=diff.new_messages[0],
                            pts=updates.pts,
                            pts_count=updates.pts_count,
                        ),
                        {u.id: u for u in diff.users},
                        {c.id: c for c in diff.chats},
                    )
                )
            else:
                if diff.other_updates:  # The other_updates list can be empty
                    self.dispatcher.updates_queue.put_nowait((diff.other_updates[0], {}, {}))
        elif isinstance(updates, raw.types.UpdateShort):
            self.dispatcher.updates_queue.put_nowait((updates.update, {}, {}))
        elif isinstance(updates, raw.types.UpdatesTooLong):
            await self.invoke(raw.functions.updates.GetState())


class TelethonUtils:
    @classmethod
    def start(
        cls,
        client: TelegramClient,
        phone: typing.Union[typing.Callable[[], str], str] = lambda: input(
            "Please enter your phone (or bot token): "
        ),
        password: typing.Union[typing.Callable[[], str], str] = lambda: getpass.getpass(
            "Please enter your password: "
        ),
        *,
        bot_token: str = None,
        force_sms: bool = False,
        code_callback: typing.Callable[[], typing.Union[str, int]] = None,
        first_name: str = "New User",
        last_name: str = "",
        max_attempts: int = 3,
    ):
        """This function is a modified version of telethon.client.auth.AuthMethods.start"""

        if code_callback is None:

            def code_callback():
                return input("Please enter the code you received: ")

        elif not callable(code_callback):
            raise ValueError(
                "The code_callback parameter needs to be a callable "
                "function that returns the code you received by Telegram."
            )

        if not phone and not bot_token:
            raise ValueError("No phone number or bot token provided.")

        if phone and bot_token and not callable(phone):
            raise ValueError("Both a phone and a bot token provided, " "must only provide one of either")

        coro = cls._telethon_start(
            client,
            phone=phone,
            password=password,
            bot_token=bot_token,
            force_sms=force_sms,
            code_callback=code_callback,
            first_name=first_name,
            last_name=last_name,
            max_attempts=max_attempts,
        )
        return coro if client.loop.is_running() else client.loop.run_until_complete(coro)

    @classmethod
    async def _telethon_start(
        cls,
        client: TelegramClient,
        phone,
        password,
        bot_token,
        force_sms,
        code_callback,
        first_name,
        last_name,
        max_attempts,
    ):
        """This function is a modified version of telethon.client.auth.AuthMethods._start"""

        from telethon import errors, utils

        if not client.is_connected():
            await client.connect()

        # Rather than using `is_user_authorized`, use `get_me`. While this is
        # more expensive and needs to retrieve more data from the server, it
        # enables the library to warn users trying to login to a different
        # account. See #1172.
        me = await client.get_me()
        if me is not None:
            # The warnings here are on a best-effort and may fail.
            if bot_token:
                # bot_token's first part has the bot ID, but it may be invalid
                # so don't try to parse as int (instead cast our ID to string).
                if bot_token[: bot_token.find(":")] != str(me.id):
                    logger.warning(
                        "该会话已经包含一个授权用户,所以没有使用提供的 bot_token 登录机器人账号(可能导致没有使用你期望的用户)"
                    )
            elif phone and not callable(phone) and utils.parse_phone(phone) != me.phone:
                logger.warning(
                    "该会话已经包含一个授权用户,所以没有使用提供的手机号登录账号 (可能导致没有使用你期望的用户)"
                )

            return client

        if not bot_token:
            # Turn the callable into a valid phone number (or bot token)
            while callable(phone):
                value = phone()
                if inspect.isawaitable(value):
                    value = await value

                if ":" in value:
                    # Bot tokens have 'user_id:access_hash' format
                    bot_token = value
                    break

                phone = utils.parse_phone(value) or phone

        if bot_token:
            await client.sign_in(bot_token=bot_token)
            return client

        me = None
        attempts = 0
        two_step_detected = False

        await client.send_code_request(phone, force_sms=force_sms)
        while attempts < max_attempts:
            try:
                value = code_callback()
                if inspect.isawaitable(value):
                    value = await value

                # Since sign-in with no code works (it sends the code)
                # we must double-check that here. Else we'll assume we
                # logged in, and it will return None as the User.
                if not value:
                    raise errors.PhoneCodeEmptyError(request=None)

                # Raises SessionPasswordNeededError if 2FA enabled
                me = await client.sign_in(phone, code=value)
                break
            except errors.SessionPasswordNeededError:
                two_step_detected = True
                break
            except (
                errors.PhoneCodeEmptyError,
                errors.PhoneCodeExpiredError,
                errors.PhoneCodeHashEmptyError,
                errors.PhoneCodeInvalidError,
            ):
                logger.warning("验证码错误, 请重试.")

            attempts += 1
        else:
            raise RuntimeError("{} consecutive sign-in attempts failed. Aborting".format(max_attempts))

        if two_step_detected:
            if not password:
                raise ValueError(
                    "Two-step verification is enabled for this account. "
                    "Please provide the 'password' argument to 'start()'."
                )

            if callable(password):
                for _ in range(max_attempts):
                    try:
                        value = password()
                        if inspect.isawaitable(value):
                            value = await value

                        me = await client.sign_in(phone=phone, password=value)
                        break
                    except errors.PasswordHashInvalidError:
                        logger.warning("两步验证密码错误, 请重试.")
                else:
                    raise errors.PasswordHashInvalidError(request=None)
            else:
                me = await client.sign_in(phone=phone, password=password)

        return client


class ClientsSession:
    pool = {}
    lock = asyncio.Lock()
    watch = None

    @classmethod
    def from_config(cls, config, in_memory=True, quiet=False, **kw):
        accounts = config.get("telegram", [])
        for k, (v, d) in kw.items():
            accounts = [a for a in accounts if a.get(k, d) in to_iterable(v)]
        return cls(
            accounts=accounts,
            proxy=config.get("proxy", None),
            basedir=config.get("basedir", None),
            in_memory=in_memory,
            quiet=quiet,
        )

    @classmethod
    async def watchdog(cls, timeout=120):
        logger.debug("Telegram 账号池看门狗启动.")
        try:
            counter = {}
            while True:
                await asyncio.sleep(10)
                for p in list(cls.pool):
                    try:
                        if cls.pool[p][1] <= 0:
                            if p in counter:
                                counter[p] += 1
                                if counter[p] >= timeout / 10:
                                    counter[p] = 0
                                    await cls.clean(p)
                            else:
                                counter[p] = 1
                        else:
                            counter.pop(p, None)
                    except (TypeError, KeyError):
                        pass
        except asyncio.CancelledError:
            await cls.shutdown()

    @classmethod
    async def clean(cls, phone):
        async with cls.lock:
            entry = cls.pool.get(phone, None)
            if not entry:
                return
            try:
                client, ref = entry
            except TypeError:
                return
            if not ref:
                logger.debug(f'登出账号 "{client.phone_number}".')
                await client.stop()
                cls.pool.pop(phone, None)

    @classmethod
    async def clean_all(cls):
        for phone in list(cls.pool):
            await cls.clean(phone)

    @classmethod
    async def shutdown(cls):
        print("\r正在停止...\r", end="", flush=True, file=sys.stderr)
        for v in cls.pool.values():
            if isinstance(v, asyncio.Task):
                v.cancel()
            else:
                client, ref = v
                client.dispatcher.updates_queue.put_nowait(None)
                for t in client.dispatcher.handler_worker_tasks:
                    try:
                        await t
                    except asyncio.CancelledError:
                        pass
        while len(asyncio.all_tasks()) > 1:
            await asyncio.sleep(0.1)
        print(f"Telegram 账号池停止.\r", end="", file=sys.stderr)
        for v in cls.pool.values():
            if isinstance(v, tuple):
                client: Client = v[0]
                await client.storage.save()
                await client.storage.close()
                logger.debug(f'登出账号 "{client.phone_number}".')

    def __init__(self, accounts, proxy=None, basedir=None, in_memory=True, quiet=False):
        self.accounts = accounts
        self.proxy = proxy
        self.basedir = basedir or user_data_dir(__product__)
        self.phones = []
        self.done = asyncio.Queue()
        self.in_memory = in_memory
        self.quiet = quiet
        if not self.watch:
            self.__class__.watch = asyncio.create_task(self.watchdog())

    def get_connector(self, proxy=None, **kw):
        if proxy:
            connector = ProxyConnector(
                proxy_type=ProxyType[proxy["scheme"].upper()],
                host=proxy["hostname"],
                port=proxy["port"],
                username=proxy.get("username", None),
                password=proxy.get("password", None),
                **kw,
            )
        else:
            connector = aiohttp.TCPConnector(**kw)
        return connector

    async def test_network(self, proxy=None):
        url = "https://www.gstatic.com/generate_204"
        connector = self.get_connector(proxy=proxy)
        async with aiohttp.ClientSession(connector=connector) as session:
            try:
                async with session.get(url) as resp:
                    if resp.status == 204:
                        return True
                    else:
                        logger.warning(f"检测网络状态时发生错误, 网络检测将被跳过.")
                        return False
            except (ProxyConnectionError, ProxyTimeoutError) as e:
                un = connector._proxy_username
                pw = connector._proxy_password
                auth = f"{un}:{pw}@" if un or pw else ""
                proxy_url = f"{connector._proxy_type.name.lower()}://{auth}{connector._proxy_host}:{connector._proxy_port}"
                logger.warning(
                    f"无法连接到您的代理 ({proxy_url}), 您的网络状态可能不好, 敬请注意. 程序将继续运行."
                )
            except OSError as e:
                logger.warning(f"无法连接到网络 (Google), 您的网络状态可能不好, 敬请注意. 程序将继续运行.")
                return False
            except Exception as e:
                logger.warning(f"检测网络状态时发生错误, 网络检测将被跳过.")
                show_exception(e)
                return False

    async def test_time(self, proxy=None):
        url = "https://timeapi.io/api/Time/current/zone?timeZone=UTC"
        connector = self.get_connector(proxy=proxy)
        async with aiohttp.ClientSession(connector=connector) as session:
            try:
                async with session.get(url) as resp:
                    if resp.status == 200:
                        resp_dict: dict = await resp.json()
                    else:
                        logger.warning(f"世界时间接口异常, 系统时间检测将跳过, 敬请注意. 程序将继续运行.")

                api_time_str = resp_dict["dateTime"]
                api_time = datetime.strptime(api_time_str.split(".")[0], "%Y-%m-%dT%H:%M:%S")
                api_time = api_time.replace(tzinfo=timezone.utc)
                api_timestamp = api_time.timestamp()

                nowtime = datetime.now(timezone.utc).timestamp()
                if abs(nowtime - api_timestamp) > 30:
                    logger.warning(
                        f"您的系统时间设置不正确, 与世界时间差距过大, 可能会导致连接失败, 敬请注意. 程序将继续运行."
                    )
            except Exception as e:
                logger.warning(f"检测世界时间发生错误, 时间检测将被跳过.")
                show_exception(e)
                return False

    @staticmethod
    async def get_session_string_from_telethon(account, proxy):
        from telethon import TelegramClient
        from telethon.sessions import StringSession

        telethon_proxy = None
        if proxy:
            telethon_proxy = {
                "proxy_type": proxy["scheme"],
                "addr": proxy["hostname"],
                "port": proxy["port"],
            }
            if "username" in proxy:
                telethon_proxy["username"] = proxy["username"]
            if "password" in proxy:
                telethon_proxy["password"] = proxy["password"]

        with tempfile.NamedTemporaryFile() as tmp_file:
            client = TelegramClient(
                tmp_file.name,
                api_id=account["api_id"],
                api_hash=account["api_hash"],
                system_version="4.16.30-vxEMBY",
                device_model="A320MH",
                app_version=__version__,
                proxy=telethon_proxy,
            )

            msg1 = f'请输入 "{account["phone"]}" 的两步验证密码 (不显示, 按回车确认)'
            password_callback = lambda: Prompt.ask(" " * 23 + msg1, password=True, console=var.console)
            msg2 = f'请输入 "{account["phone"]}" 的登陆验证码 (按回车确认)'
            code_callback = lambda: Prompt.ask(" " * 23 + msg2, console=var.console)

            try:
                await TelethonUtils.start(
                    client,
                    phone=account["phone"],
                    password=password_callback,
                    code_callback=code_callback,
                )
                session_string = StringSession.save(client.session)
                me = await client.get_me()
                user_id = me.id
                user_bot = me.bot
            except RuntimeError:
                return None
            finally:
                await client.disconnect()

        session = StringSession(session_string)
        Dt = Storage.SESSION_STRING_FORMAT
        return (
            base64.urlsafe_b64encode(
                struct.pack(
                    Dt,
                    session.dc_id,
                    int(account["api_id"]),
                    None,
                    session.auth_key.key,
                    user_id,
                    user_bot,
                )
            )
            .decode()
            .rstrip("=")
        )

    async def login(self, index, account, proxy, use_telethon=True):
        try:
            account["phone"] = "".join(account["phone"].split())
            Path(self.basedir).mkdir(parents=True, exist_ok=True)
            session_file = Path(self.basedir) / f'{account["phone"]}.session'
            session_string_file = Path(self.basedir) / f'{account["phone"]}.login'
            if not self.quiet:
                logger.info(f'登录至账号 "{account["phone"]}".')
            for _ in range(3):
                if account.get("api_id", None) is None or account.get("api_hash", None) is None:
                    account.update(random.choice(list(API_KEY.values())))
                config_session_string = session_string = account.get("session", None)
                file_session_string = None
                if not session_string:
                    if session_string_file.is_file():
                        with open(session_string_file, encoding="utf-8") as f:
                            file_session_string = session_string = f.read().strip()
                if self.in_memory is None:
                    in_memory = True
                    if not session_string:
                        if session_file.is_file():
                            in_memory = False
                elif session_string:
                    in_memory = True
                else:
                    in_memory = self.in_memory
                if session_string or session_file.is_file():
                    logger.debug(
                        f'账号 "{account["phone"]}" 登录凭据存在, 仅内存模式{"启用" if in_memory else "禁用"}.'
                    )
                else:
                    logger.debug(
                        f'账号 "{account["phone"]}" 登录凭据不存在, 即将进入登录流程, 仅内存模式{"启用" if in_memory else "禁用"}.'
                    )
                    if use_telethon:
                        logger.debug("选择使用 Telethon 进行首次登陆, 并转发字符串至 Pyrogram.")
                        file_session_string = session_string = await self.get_session_string_from_telethon(
                            account, proxy
                        )
                        if session_string:
                            logger.info("请耐心等待, 正在登陆.")
                            await asyncio.sleep(10)
                        else:
                            logger.warning(f'登录账号 "{account["phone"]}" 尝试次数超限, 将被跳过.')
                            return None
                try:
                    client = Client(
                        app_version=__version__,
                        device_model="A320MH",
                        name=account["phone"],
                        system_version="4.16.30-vxEmby",
                        api_id=account["api_id"],
                        api_hash=account["api_hash"],
                        phone_number=account["phone"],
                        session_string=session_string,
                        in_memory=in_memory,
                        proxy=proxy,
                        workdir=self.basedir,
                        sleep_threshold=30,
                    )
                    try:
                        await asyncio.wait_for(client.start(), 120)
                    except asyncio.TimeoutError:
                        if proxy:
                            logger.error(f"无法连接到 Telegram 服务器, 请检查您代理的可用性.")
                            continue
                        else:
                            logger.error(f"无法连接到 Telegram 服务器, 请检查您的网络.")
                            continue
                except OperationalError as e:
                    logger.warning(f"内部数据库错误, 正在重置, 您可能需要重新登录.")
                    show_exception(e)
                    session_file.unlink(missing_ok=True)
                except ApiIdPublishedFlood:
                    logger.warning(f'登录账号 "{account["phone"]}" 时发生 API key 限制, 将被跳过.')
                    break
                except Unauthorized as e:
                    if config_session_string:
                        logger.error(
                            f'账号 "{account["phone"]}" 由于配置中提供的 session 已被注销, 将被跳过.'
                        )
                        show_exception(e)
                        break
                    elif file_session_string:
                        logger.error(f'账号 "{account["phone"]}" 已被注销, 将在 3 秒后重新登录.')
                        show_exception(e)
                        session_string_file.unlink(missing_ok=True)
                        continue
                    elif client.in_memory:
                        logger.error(f'账号 "{account["phone"]}" 已被注销, 将在 3 秒后重新登录.')
                        show_exception(e)
                        continue
                    else:
                        logger.error(f'账号 "{account["phone"]}" 已被注销, 将在 3 秒后重新登录.')
                        show_exception(e)
                        await client.storage.delete()
                except KeyError as e:
                    logger.warning(
                        f'登录账号 "{account["phone"]}" 时发生异常, 可能是由于网络错误, 将在 3 秒后重试.'
                    )
                    show_exception(e)
                    await asyncio.sleep(3)
                else:
                    break
            else:
                logger.error(f'登录账号 "{account["phone"]}" 失败次数超限, 将被跳过.')
                return None
        except asyncio.CancelledError:
            raise
        except binascii.Error:
            logger.error(
                f'登录账号 "{account["phone"]}" 失败, 由于您在配置文件中提供的 session 无效, 将被跳过.'
            )
        except RPCError as e:
            logger.error(f'登录账号 "{account["phone"]}" 失败 ({e.MESSAGE.format(value=e.value)}), 将被跳过.')
            return None
        except BadMsgNotification as e:
            if "synchronized" in str(e):
                logger.error(
                    f'登录账号 "{account["phone"]}" 时发生异常, 可能是因为您的系统时间与世界时间差距过大, 将被跳过.'
                )
                return None
            else:
                logger.error(f'登录账号 "{account["phone"]}" 时发生异常, 将被跳过.')
                show_exception(e, regular=False)
                return None
        except Exception as e:
            logger.error(f'登录账号 "{account["phone"]}" 时发生异常, 将被跳过.')
            show_exception(e, regular=False)
            return None
        else:
            if not session_string_file.exists():
                with open(session_string_file, "w+", encoding="utf-8") as f:
                    f.write(await client.export_session_string())
            logger.debug(f'登录账号 "{client.phone_number}" 成功.')
            client._login_time = datetime.now()
            client._config_index = index
            return client

    async def loginer(self, index, account):
        client = await self.login(index, account, proxy=self.proxy)
        if isinstance(client, Client):
            async with self.lock:
                phone = account["phone"]
                self.pool[phone] = (client, 1)
                self.phones.append(phone)
                await self.done.put(client)
                logger.debug(f"Telegram 账号池计数增加: {phone} => 1")
        else:
            await self.done.put(None)

    async def __aenter__(self):
        await self.test_network(self.proxy)
        asyncio.create_task(self.test_time(self.proxy))
        for i, a in enumerate(self.accounts):
            phone = a["phone"].replace(" ", "")
            try:
                await self.lock.acquire()
                if phone in self.pool:
                    if isinstance(self.pool[phone], asyncio.Task):
                        self.lock.release()
                        await self.pool[phone]
                        await self.lock.acquire()
                    if isinstance(self.pool[phone], asyncio.Task):
                        continue
                    client, ref = self.pool[phone]
                    ref += 1
                    self.pool[phone] = (client, ref)
                    self.phones.append(phone)
                    await self.done.put(client)
                    logger.debug(f"Telegram 账号池计数增加: {phone} => {ref}")
                else:
                    self.pool[phone] = asyncio.create_task(self.loginer(i, a))
            finally:
                try:
                    self.lock.release()
                except RuntimeError:
                    pass
        return self

    def __aiter__(self):
        async def aiter():
            for _ in range(len(self.accounts)):
                client: Client = await self.done.get()
                if client:
                    yield client

        return aiter()

    async def __aexit__(self, type, value, tb):
        async with self.lock:
            for phone in self.phones:
                entry = self.pool.get(phone, None)
                if entry:
                    client, ref = entry
                    ref -= 1
                    self.pool[phone] = (client, ref)
                    logger.debug(f"Telegram 账号池计数降低: {phone} => {ref}")



from pyrogram.types import Message

from ._base import AnswerBotCheckin

__ignore__ = True


class TembyCheckin(AnswerBotCheckin):
    name = "Temby"
    bot_username = "HiEmbyBot"
    bot_checkin_cmd = "/hi"
    bot_success_keywords = ["Checkin successfully"]
    bot_checked_keywords = ["you have checked in already today"]

    async def on_answer(self, message: Message):
        await super().on_answer(message)
        keys = [k.text for r in message.reply_markup.inline_keyboard for k in r]
        if len(keys) == 1:
            await message.click(k)
        else:
            for k in keys:
                if "签到" in k:
                    await message.click(k)
                    return
            else:
                self.log.warning(f"签到失败: 账户错误.")
                return await self.fail()



import asyncio
import csv
from datetime import datetime
from pathlib import Path
import random

from pyrogram.types import Message
from pyrogram.errors import RPCError

from ...utils import to_iterable, truncate_str
from ..link import Link
from ..lock import pornemby_alert
from ._base import Monitor


class _PornembyAnswerResultMonitor(Monitor):
    name = "Pornemby 科举答案"
    chat_keyword = r"问题\d*：(.*?)\n+A:(.*)\n+B:(.*)\n+C:(.*)\n+D:(.*)\n+答案为：([ABCD])"
    additional_auth = ["pornemby_pack"]
    allow_edit = True

    key_map = {"A": 1, "B": 2, "C": 3, "D": 4}

    async def on_trigger(self, message: Message, key, reply):
        spec = f"[gray50]({truncate_str(key[0], 10)})[/]"
        self.log.info(f"本题正确答案为 {key[5]} ({key[self.key_map[key[5]]]}): {spec}.")


class _PornembyAnswerAnswerMonitor(Monitor):
    name = "Pornemby 科举"
    history_chat_name = "Pornemby"
    chat_user = [
        "pornemby_question_bot",
        "PronembyTGBot2_bot",
        "PronembyTGBot3_bot",
        "PornembyBot",
        "Porn_Emby_Bot",
    ]
    chat_keyword = r"问题\d*：(.*?)\n+(A:.*\n+B:.*\n+C:.*\n+D:.*)\n(?!\n*答案)"
    additional_auth = ["pornemby_pack"]
    allow_edit = True

    cache = {}
    lock = asyncio.Lock()

    key_map = {
        "A": "🅰",
        "B": "🅱",
        "C": "🅲",
        "D": "🅳",
    }

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        self.cache_file = Path(self.basedir) / "pornemby_question.csv"
        self.update_task = None

    async def update_cache(self, to_date=None):
        cache_timestamp = self.cache_file.with_name("pornemby_question.timestamp")
        self.cache_file.parent.mkdir(parents=True, exist_ok=True)
        if not self.cache_file.exists():
            created = True
            to_date = datetime.fromtimestamp(0)
            self.log.info("首次使用 Pornemby 科举, 正在缓存问题答案历史.")
        else:
            created = False
            if cache_timestamp.exists() and not to_date:
                with open(cache_timestamp) as f:
                    try:
                        to_date = datetime.fromtimestamp(float(f.read()))
                    except ValueError:
                        to_date = datetime.fromtimestamp(0)
            else:
                to_date = datetime.fromtimestamp(0)
            self.log.info(f"正在更新问题答案历史缓存.")
            self.log.debug(f"上一次问题答案历史写入于 {to_date.strftime('%Y-%m-%d %H:%M')}.")
        count = 0
        qs = 0
        finished = False
        with open(self.cache_file, mode="a+", newline="", encoding="utf-8") as csvfile:
            writer = csv.writer(csvfile)
            if created:
                writer.writerow(["Question", "A", "B", "C", "D", "Answer"])
            while not finished:
                finished = True
                m: Message
                for g in to_iterable(self.history_chat_name):
                    async for m in self.client.search_messages(g, limit=100, offset=count, query="答案为"):
                        if m.date < to_date:
                            break
                        count += 1
                        finished = False
                        if m.text:
                            for key in _PornembyAnswerResultMonitor.keys(m):
                                qs += 1
                                writer.writerow(key)
                if count and (finished or count % 500 == 0):
                    self.log.info(f"读取问题答案历史: 已读取 {qs} 问题 / {count} 信息.")
                    await asyncio.sleep(2)
        self.log.debug(f"已向问题答案历史缓存写入 {qs} 条问题.")
        with open(cache_timestamp, "w+", encoding="utf-8") as f:
            f.write(str(datetime.now().timestamp()))

    async def read_cache(self):
        if not self.cache_file.exists():
            return {}
        questions = {}
        with open(self.cache_file, "r", encoding="utf-8") as csvfile:
            rows = csv.DictReader(csvfile)
            for row in rows:
                entry = dict(row)
                questions[entry["Question"]] = entry["Answer"]
        self.log.debug(f"已从问题答案历史缓存读取 {len(questions)} 条问题.")
        return questions

    async def update(self):
        try:
            await asyncio.wait_for(self.lock.acquire(), 1)
        except asyncio.TimeoutError:
            self.log.debug("等待其他协程缓存问题答案历史.")
            async with self.lock:
                return True
        else:
            try:
                await self.update_cache()
                self.__class__.cache = await self.read_cache()
                return True
            finally:
                self.lock.release()

    async def cache_watchdog(self):
        try:
            while True:
                secs = 3600 * 12
                self.log.debug(f"等待 {secs} 秒后进行缓存更新.")
                await asyncio.sleep(secs)
                await self.update()
        except asyncio.CancelledError:
            raise

    async def init(self):
        self.update_task = asyncio.create_task(self.cache_watchdog())
        return await self.update()

    async def on_trigger(self, message: Message, key, reply):
        spec = f"[gray50]({truncate_str(key[0], 10)})[/]"
        if pornemby_alert.get(self.client.me.id, False):
            self.log.info(f"由于风险急停不作答: {spec}.")
            return
        if random.random() > self.config.get("possibility", 1.0):
            self.log.info(f"由于概率设置不作答: {spec}.")
            return
        result = self.cache.get(key[0], None)
        if result:
            self.log.info(f"从缓存回答问题为{result}: {spec}.")
        elif self.config.get("only_history", False):
            self.log.info(f"未从历史缓存找到问题, 请自行回答: {spec}.")
        else:
            for retries in range(3):
                self.log.debug(f"未从历史缓存找到问题, 开始请求云端问题回答: {spec}.")
                result, by = await Link(self.client).pornemby_answer(key[0] + "\n" + key[1])
                if result:
                    self.log.info(f"请求 {by or '云端'} 问题回答为{result}: {spec}.")
                    break
                else:
                    self.log.info(f"云端问题回答错误或超时, 正在重试: {spec}.")
            else:
                self.log.info(f"错误次数超限, 回答失败: {spec}.")
                return
        try:
            await asyncio.sleep(random.uniform(2, 4))
            answer = await message.click(self.key_map[result])
            self.log.debug(f"回答返回值: {answer.message} {spec}.")
        except KeyError:
            self.log.info(f"点击失败: {result} 不是可用的答案 {spec}.")
        except RPCError:
            self.log.info(f"点击失败: 问题已失效.")


class PornembyAnswerMonitor:
    class PornembyAnswerResultMonitor(_PornembyAnswerResultMonitor):
        chat_name = "Pornemby"

    class PornembyAnswerAnswerMonitor(_PornembyAnswerAnswerMonitor):
        chat_name = "Pornemby"



from ._base import BotCheckin

__ignore__ = True


class AkileGroupCheckin(BotCheckin):
    name = "Akile 群组发言"
    chat_name = "akileChat"
    bot_checkin_cmd = ["/checkin@akilecloud_bot"]
    bot_account_fail_keywords = ["未绑定"]



from ._base import BotCheckin

__ignore__ = True


class PinganSGKCheckin(BotCheckin):
    name = "平安社工库"
    bot_username = "pingansgk_bot"
    bot_checkin_cmd = "/qd"
    additional_auth = ["prime"]



from ._base import BotCheckin

__ignore__ = True


class AISGKCheckin(BotCheckin):
    name = "AI 社工库"
    bot_username = "aishegongkubot"
    bot_checkin_cmd = "/sign"
    additional_auth = ["prime"]
    bot_checked_keywords = "请勿重复签到"



from ._base import BotCheckin

__ignore__ = True


class StarSGKCheckin(BotCheckin):
    name = "星月社工库"
    bot_username = "XY_SGKBOT"
    bot_checkin_cmd = "/sign"
    additional_auth = ["prime"]
    checked_retries = 6



from ._base import BotCheckin


class DataSGKCheckin(BotCheckin):
    name = "数据社工库"
    bot_username = "DataSGK_bot"
    bot_checkin_cmd = "/signin"
    additional_auth = ["prime"]
    bot_checked_keywords = "请勿重复签到"



from pyrogram.types import Message

from ._base import Monitor

__ignore__ = True


class EmbyhubMonitor(Monitor):
    name = "EmbyHub"
    chat_name = "emby_hub"
    chat_user = "ednovas"
    chat_keyword = r"注册已开放"
    bot_username = "EdHubot"
    notify_create_name = True
    additional_auth = ["prime"]

    async def on_trigger(self, message: Message, key, reply):
        await self.client.send_message(self.bot_username, f"/create {self.unique_name}")
        self.log.bind(msg=True).info(
            f'已向 Bot @{self.bot_username} 发送了用户注册申请: "{self.unique_name}", 请检查结果.'
        )



from ._base import BotCheckin

__ignore__ = True


class TestGroupCheckin(BotCheckin):
    name = "群组签到测试"
    chat_name = "api_group"
    bot_checkin_cmd = "签到"

    async def send_checkin(self, retry=False):
        super().send_checkin(retry)
        self.finished.set()



from ._base import BotCheckin

__ignore__ = True


class TianmaoSGKCheckin(BotCheckin):
    name = "天猫社工库"
    bot_username = "UISGKbot"
    bot_checkin_cmd = "/sign"
    additional_auth = ["prime"]
    checked_retries = 6



from pyrogram.types import Message
from pyrogram.raw.types.messages import BotCallbackAnswer

from embykeeper.utils import to_iterable

from ._base import BotCheckin

__ignore__ = True


class TemplateACheckin(BotCheckin):
    bot_checkin_cmd = "/start"
    templ_panel_keywords = ["请选择功能", "用户面板"]
    use_button_answer = True
    bot_text_ignore_answer = ["Done"]

    async def message_handler(self, client, message: Message):
        text = message.caption or message.text
        if (
            text
            and any(keyword in text for keyword in to_iterable(self.templ_panel_keywords))
            and message.reply_markup
        ):
            keys = [k.text for r in message.reply_markup.inline_keyboard for k in r]
            for k in keys:
                if "签到" in k or "簽到" in k:
                    try:
                        answer: BotCallbackAnswer = await message.click(k)
                    except TimeoutError:
                        self.log.debug(f"点击签到按钮无响应, 可能按钮未正确处理点击回复. 一般来说不影响签到.")
                    else:
                        if self.use_button_answer:
                            if not any(ignore in answer.message for ignore in self.bot_text_ignore_answer):
                                await self.on_text(Message(id=0, text=answer.message), answer.message)
                    return
            else:
                self.log.warning(f"签到失败: 账户错误.")
                return await self.fail()

        if message.text and "请先点击下面加入我们的" in message.text:
            self.log.warning(f"签到失败: 账户错误.")
            return await self.fail()

        await super().message_handler(client, message)


def use(**kw):
    return type("TemplatedClass", (TemplateACheckin,), kw)



from ._base import BotCheckin

__ignore__ = True


class KoiSGKCheckin(BotCheckin):
    name = "Koi 社工库"
    bot_username = "KoiSGKbot"
    bot_checkin_cmd = "/qd"
    additional_auth = ["prime"]
    bot_checked_keywords = ["今日已签到"]



from ._base import BotCheckin

__ignore__ = True


class HuaSGKCheckin(BotCheckin):
    name = "花花社工库"
    bot_username = "sgkvipbot"
    bot_checkin_cmd = "/sign"
    additional_auth = ["prime"]
    checked_retries = 6



import asyncio
import random
import string

from pyrogram.types import Message
from PIL import Image

from ...utils import async_partial, nonblocking
from ...data import get_datas
from ..lock import misty_locks
from ._base import Monitor

misty_monitor_pool = {}


class MistyMonitor(Monitor):
    ocr = "digit5-large@v1"
    ocr_lock = asyncio.Lock()

    name = "Misty"
    chat_name = "FreeEmbyGroup"
    chat_keyword = r"空余名额数: (?!0$)"
    bot_username = "EmbyMistyBot"
    notify_create_name = True
    additional_auth = ["prime"]

    async def init(self, initial=True, force_lock=True):
        from ddddocr import DdddOcr

        async with self.ocr_lock:
            if isinstance(self.ocr, str):
                data = []
                files = (f"{self.ocr}.onnx", f"{self.ocr}.json")
                async for p in get_datas(self.basedir, files, proxy=self.proxy, caller=self.name):
                    if p is None:
                        self.log.info(f"初始化错误: 无法下载所需文件.")
                        return False
                    else:
                        data.append(p)
                self.__class__.ocr = DdddOcr(
                    show_ad=False, import_onnx_path=str(data[0]), charsets_path=str(data[1])
                )

        misty_monitor_pool[self.client.me.id] = self
        self.captcha = None
        self.log.info(f"正在初始化机器人状态.")
        wr = async_partial(self.client.wait_reply, self.bot_username)
        misty_locks.setdefault(self.client.me.id, asyncio.Lock())
        lock = misty_locks.get(self.client.me.id, None)
        if not force_lock:
            lock = nonblocking(lock)
        async with lock:
            for _ in range(20 if initial else 3):
                try:
                    msg: Message = await wr("/cancel")
                    if "选择您要使用的功能" in (msg.caption or msg.text):
                        await asyncio.sleep(random.uniform(2, 4))
                        msg = await wr("⚡️账号功能")
                    if "请选择功能" in (msg.text or msg.caption):
                        await asyncio.sleep(random.uniform(2, 4))
                        msg = await wr("⚡️注册账号")
                        if "请输入验证码" in (msg.caption or msg.text):
                            data = await self.client.download_media(msg, in_memory=True)
                            image = Image.open(data)
                            self.captcha = (
                                self.ocr.classification(image)
                                .translate(str.maketrans("", "", string.punctuation))
                                .replace(" ", "")
                            )
                            self.log.debug(f"接收到验证码: {self.captcha}")
                except (asyncio.TimeoutError, TypeError):
                    continue
                else:
                    if self.captcha and len(self.captcha) == 5:
                        self.log.info(
                            f"机器人状态初始化完成, 当接收到邀请码时将输入验证码 {self.captcha} 以抢注, 请勿操作 @EmbyMistyBot."
                        )
                        return True
                    else:
                        self.log.info(f"机器人状态初始化失败, 正在重试.")
            else:
                self.log.bind(log=True).warning(f"Misty 机器人状态初始化失败, 监控将停止.")
                return False

    async def on_trigger(self, message: Message, key, reply):
        wr = async_partial(self.client.wait_reply, self.bot_username)
        misty_locks.setdefault(self.client.me.id, asyncio.Lock())
        lock = misty_locks.get(self.client.me.id, None)
        async with lock:
            for _ in range(3):
                try:
                    msg = await wr(self.captcha)
                    if "验证码错误" in msg.text:
                        self.log.info(f"验证码错误, 将重新初始化.")
                        if not await self.init(force_lock=False):
                            self.failed.set()
                    elif "暂时停止注册" in msg.text:
                        self.log.info(f"注册名额已满, 将重新初始化并继续监控.")
                        if not await self.init(force_lock=False):
                            self.failed.set()
                        return
                    elif "用户名" in msg.text:
                        msg = await wr(self.unique_name)
                        if "密码" in msg.text:
                            await self.client.send_message(self.bot_username, "/cancel")
                            self.log.bind(msg=True).info(
                                f'已向 Bot @{self.bot_username} 发送了用户注册申请: "{self.unique_name}", 请检查结果.'
                            )
                except asyncio.TimeoutError:
                    pass
            else:
                self.log.info(f"未成功, 结束注册申请.")



from ._base import BotCheckin

__ignore__ = True


class MasterSGKCheckin(BotCheckin):
    name = "Master 社工库"
    bot_username = "BaKaMasterBot"
    bot_checkin_cmd = "/sign"
    additional_auth = ["prime"]
    checked_retries = 6



from pyrogram.types import Message

from embykeeper.utils import to_iterable

from ._base import AnswerBotCheckin, MessageType


class TanhuaCheckin(AnswerBotCheckin):
    name = "探花"
    bot_username = "TanhuaTvBot"
    additional_auth = ["prime"]
    bot_checkin_cmd = ["/start"]
    templ_panel_keywords = ["请选择功能", "用户面板", "用户名称"]
    bot_success_keywords = ["签到获得积分"]
    bot_checked_keywords = ["今日已签到"]
    bot_text_ignore = ["请选择正确的验证码"]

    async def message_handler(self, client, message: Message):
        text = message.caption or message.text
        if (
            text
            and any(keyword in text for keyword in to_iterable(self.templ_panel_keywords))
            and message.reply_markup
        ):
            keys = [k.text for r in message.reply_markup.inline_keyboard for k in r]
            for k in keys:
                if "个人信息" in k:
                    try:
                        await message.click(k)
                    except TimeoutError:
                        pass
                    return
                if "签到" in k or "簽到" in k:
                    try:
                        await message.click(k)
                    except TimeoutError:
                        self.log.debug(f"点击签到按钮无响应, 可能按钮未正确处理点击回复. 一般来说不影响签到.")
                    return
            else:
                self.log.warning(f"签到失败: 账户错误.")
                return await self.fail()

        if message.text and "请先加入聊天群组和通知频道" in message.text:
            self.log.warning(f"签到失败: 账户错误.")
            return await self.fail()

        await super().message_handler(client, message)

    def _message_type(self, message: Message):
        if message.photo:
            if message.caption:
                return MessageType.CAPTION
            else:
                return MessageType.CAPTCHA
        elif message.text:
            return MessageType.TEXT

    def message_type(self, message: Message):
        if self.is_valid_answer(message):
            return MessageType.ANSWER | self._message_type(message)
        else:
            return self._message_type(message)



from pyrogram.types import Message
from pyrogram.errors import RPCError

from ..lock import pornemby_alert
from ._base import Monitor

__ignore__ = True


class PornembyRegisterMonitor(Monitor):
    name = "Pornemby 抢注"
    chat_name = "Pornemby"
    chat_user = "PornembyTGBot_bot"
    chat_keyword = "开 放 注 册"
    additional_auth = ["pornemby_pack"]

    async def on_trigger(self, message: Message, key, reply):
        if pornemby_alert.get(self.client.me.id, False):
            self.log.info(f"由于风险急停不抢注.")
            return
        try:
            await message.click(0)
        except TimeoutError:
            self.log.info("检测到 Pornemby 抢注, 已点击, 请自行查看结果.")
        except RPCError:
            self.log.info("检测到 Pornemby 抢注, 已点击, 请自行查看结果.")
        else:
            self.log.info("检测到 Pornemby 抢注, 已点击, 请自行查看结果.")



import aiohttp
from aiohttp_socks import ProxyConnector, ProxyType
from pyrogram.types import Message
from pyrogram.raw.functions.messages import AcceptUrlAuth
from pyrogram.raw.types import UrlAuthResultAccepted
from faker import Faker

from ._base import BotCheckin

__ignore__ = True


class JMSCheckin(BotCheckin):
    name = "卷毛鼠"
    bot_username = "jmsembybot"
    bot_checked_keywords = "请明天再来签到"

    async def message_handler(self, client, message: Message):
        if message.reply_markup:
            keys = [k for r in message.reply_markup.inline_keyboard for k in r]
            for k in keys:
                if "点我签到" in k.text:
                    r: UrlAuthResultAccepted = await self.client.invoke(
                        AcceptUrlAuth(
                            peer=await self.client.resolve_peer(message.chat.id),
                            msg_id=message.id,
                            button_id=k.login_url.button_id,
                            url=k.login_url.url,
                        )
                    )
                    url = r.url
                    if self.proxy:
                        connector = ProxyConnector(
                            proxy_type=ProxyType[self.proxy["scheme"].upper()],
                            host=self.proxy["hostname"],
                            port=self.proxy["port"],
                            username=self.proxy.get("username", None),
                            password=self.proxy.get("password", None),
                        )
                    else:
                        connector = aiohttp.TCPConnector()
                    for _ in range(1, 3):
                        async with aiohttp.ClientSession(connector=connector) as session:
                            async with session.get(url, headers={"User-Agent": Faker().safari()}) as resp:
                                if resp.status == 200:
                                    return
        return await super().message_handler(client, message)
